# -*- coding: utf-8 -*-
"""
@file
@brief Modules for Orange
"""
from ..installhelper.module_install import ModuleInstall


def orange_set():
    """
    modules implemented for Orange, it requires the modules in set *ml*
    """
    mod = [
        ModuleInstall(
            "pdfminer3k", "pip",
            purpose="PDF parser and analyzer"),
        ModuleInstall(
            "docx2txt", "pip",
            purpose="A pure python-based utility to extract text and images from docx files."),
        ModuleInstall(
            "odfpy", "pip",
            purpose="Python API and tools to manipulate OpenDocument files"),
        ModuleInstall(
            "setuptools-git", "pip", mname="setuptools_git",
            purpose="This is a plugin for setuptools that enables git integration. Once installed, Setuptools " +
            "can be told to include in a package distribution all the files tracked by git. " +
            "This is an alternative to explicit inclusion specifications with MANIFEST.in."),
        ModuleInstall(
            "scikit-fusion", "pip", mname="skfusion",
            purpose="A Python module for data fusion built on top of factorized models."),
        ModuleInstall(
            "AnyQt", "pip",
            purpose="PyQt4/PyQt5 compatibility layer."),
        ModuleInstall("hyperframe", "pip",
                      purpose="HTTP/2 framing layer for Python"),
        ModuleInstall("hpack", "pip",
                      purpose="Pure-Python HPACK header compression"),
        ModuleInstall("h2", "pip",
                      purpose="HTTP/2 State-Machine based protocol implementation"),
        ModuleInstall("hyper", "pip",
                      purpose="HTTP/2 Client for Python"),
        ModuleInstall("serverfiles", "pip",
                      purpose="An utility that accesses files on a HTTP server and stores them locally for reuse."),
        ModuleInstall(
            "orange3", "pip", mname="Orange",
            usage="ORANGE", purpose="Orange is a component-based data mining software. It includes a range of " +
            "data visualization, exploration, " +
            "preprocessing and modeling techniques. It can be used through a nice and intuitive user interface or, for more advanced users, " +
            "as a module for the Python programming language."),
        ModuleInstall("ufal.udpipe", "pip", purpose="for orange3-text"),
        ModuleInstall(
            "orange3-text", "pip", mname="orangecontrib.text", usage="ORANGE",
            purpose="Orange3 Text extends Orange3, a data mining software package, with common functionality for text mining. " +
            "It provides access to publicly available data, like NY Times, Twitter and PubMed. Further, " +
            "it provides tools for preprocessing, " +
            "constructing vector spaces (like bag-of-words, topic modeling and word2vec) and visualizations like word cloud end geo map. " +
            "All features can be combined with powerful data mining techniques from the Orange data mining framework."),
        ModuleInstall(
            "orange3-associate", "pip", mname="orangecontrib.associate", usage="ORANGE",
            purpose="This module implements FP-growth [1] frequent pattern mining algorithm with bucketing " +
            "optimization [2] for conditional databases of few items."),
        # has a dependency on Orange (Python 2)
        # ModuleInstall(
        #     "orange3-datafusion", "pip", mname="orangecontrib.datafusion", usage="ORANGE",
        #     purpose="This is a data fusion add-on for [Orange3](http://orange.biolab.si). Add-on wraps scikit-fusion, " +
        #     "a Python library for data fusion, and implements a set of widgets for loading of the data, definition of data fusion schema, " +
        #     "collective matrix factorization and exploration of latent factors."),
        # ModuleInstall(
        #     "orange3-network", "github", "biolab", mname="orangecontrib.network", custom=["build", "install"], usage="ORANGE",
        # purpose="Orange Network is an add-on for Orange data mining software
        # package. It provides network visualization and analysis tools."),
        ModuleInstall(
            "Orange3-Network", "pip", mname="orangecontrib.network", usage="ORANGE",
            purpose="Orange Network is an add-on for Orange data mining software package. It provides " +
            "network visualization and analysis tools."),
        ModuleInstall(
            "CacheControl", "pip", mname="cachecontrol",
            purpose="httplib2 caching for requests"),
        ModuleInstall(
            "Orange3-ImageAnalytics", "pip", mname="orangecontrib.imageanalytics", usage="ORANGE",
            purpose="Orange3 add-on for image data mining."),
        # weird
        # ModuleInstall(
        #     "orange3-prototypes", "pip", mname="orangecontrib.prototypes", usage="ORANGE",
        #     purpose="Prototype Orange widgets. Only for the brave."),
        # weird
        # ModuleInstall(
        #     "orange3-spark", "pip", mname="orangecontrib.spark", usage="ORANGE",
        # purpose="A set of widgets for Orange data mining suite to work with
        # Apache Spark ML API."),
    ]
    #
    return [_ for _ in mod if _ is not None]



"""
@file
@brief To install packages for a specific distribution.
"""
from __future__ import print_function

import os
import sys
from .win_exception import WinInstallPackageException
from ..installhelper.install_cmd_helper import run_cmd, get_pip_program
from ..packaged import small_set, find_module_install

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


def get_modules_version(python_path):
    """
    return a dictionary { module:version }

    @param      python_path     path to python
    @return                     dictionary
    """
    if sys.platform.startswith("win"):
        prog = os.path.join(python_path, "Scripts", "pip.exe")
        if not os.path.exists(prog):
            prog = get_pip_program(exe=python_path)
    else:
        prog = get_pip_program(exe=python_path)

    cmd = prog + " list"
    from pip import __version__
    if int(__version__.split(".")[0]) >= 9:
        cmd += " --format=columns"

    try:
        out, err = run_cmd(cmd, wait=True, fLOG=None, change_path=python_path)
    except Exception as e:
        raise Exception("unable to run: {0}".format(cmd)) from e

    if err is not None and len(err) > 0:
        if len(err.split("\n")) > 3 or \
           "You should consider upgrading via the 'pip install --upgrade pip' command." not in err:
            raise Exception("unable to run, #lines {0}\nERR-8:\n{1}\nOUT:\n{2}".format(
                len(err.split("\n")), err, out))

    lines = out.split("\n")
    res = {}
    for line in lines:
        if "." in line:
            spl = line.split()
            if len(spl) == 2:
                a = spl[0]
                b = spl[1].strip(" \n\r")
                res[a] = b.strip("()")
                al = a.lower()
                if al != a:
                    res[al] = res[a]
    return res


def win_install_package_other_python(python_path, package, verbose=False, deps=True, fLOG=print):
    """
    Install a package for another Python distribution than the current one.

    @param      python_path     location of python
    @param      package         location of the package (.tar.gz, .whl, .tgz, .bz2)
    @param      verbose         display more information
    @param      deps            take dependencies into account or not
    @param      fLOG            logging function
    @return                     operations ("pip", module) if installed, empty if already installed

    .. versionchanged:: 1.1
        ``deps=False`` is the default for module zipline
    """
    thename = os.path.split(package)[-1]
    if thename == "jenkins.zip":
        # jenkins.zip is not a python package
        return []
    if verbose:
        fLOG("[pymy] *** INSTALL", package)
    operations = []

    if sys.version_info[0] == 2:
        pip = os.path.join(python_path, "Scripts", "pip.exe")
    else:
        pip = os.path.join(python_path, "Scripts",
                           "pip%d.exe" % sys.version_info[0])

    if not os.path.exists(pip):
        raise FileNotFoundError(pip)

    cmd = "{0} install {1}".format(pip, package)
    if verbose:
        fLOG(cmd)

    cur = os.getcwd()
    if cur != python_path:
        os.chdir(python_path)
    name = os.path.split(package)[-1]
    if (deps is not None and not deps) or name.startswith("zipline"):
        cmd += " --no-deps"
    out, err = run_cmd(cmd, wait=True, fLOG=fLOG)

    if cur != python_path:
        os.chdir(cur)

    if verbose:
        fLOG(out)

    if err is not None and len(err) > 0:
        name = "-".join(package.split("-")[:-1])
        look = "Successfully installed " + name
        if look not in err:
            raise WinInstallPackageException(
                "unable to install {0}, due to:\nCMD\n{3}\nOUT:\n{1}\nERR-9:\n{2}\nNOT FOUND\n{4}".format(package, out, err, cmd, look))

    if "No distributions matching the version" in out:
        raise WinInstallPackageException(
            "unable to install " +
            package +
            "\nCMD\n" +
            cmd +
            "\nOUT:\n" +
            out +
            "\nERR--A:\n" +
            err)
    elif "Testing of typecheck-decorator passed without failure." in out:
        operations.append(("pip", package))
    elif "Successfully installed" not in out:
        if "error: Unable to find vcvarsall.bat" in out:
            url = "http://www.xavierdupre.fr/blog/2013-07-07_nojs.html"
            raise WinInstallPackageException(
                "unable to install " +
                package +
                "\nread:\n" +
                url +
                "\nCMD\n" +
                cmd +
                "OUT:\n" +
                out +
                "\nERR--B:\n" +
                err)
        if "Requirement already satisfied" not in out:
            raise WinInstallPackageException(
                "unable to install " +
                package +
                "\nCMD\n" +
                cmd +
                "\nOUT:\n" +
                out +
                "\nERR--C:\n" +
                err)
    else:
        operations.append(("pip", package))

    return operations


def _is_package_in_list(module_name, list_packages, no_wheel=False):
    """
    determines of this package is the one for the given module_name

    @param      list_packages       list of packages names (list of wheel)
    @param      module_name         module name
    @param      no_wheel            skip wheels
    @return                         package name
    """
    module_name = module_name.lower()
    p = "." in module_name
    if p:
        pp = module_name.replace(".", "_")

    d = "-" in module_name
    if d:
        pd = module_name.split("-")
        if len(pd[-1]) == 0:
            pd = "_".join(pd[:-1]) + "-"
        else:
            pd = "_".join(pd)

    for a in list_packages:
        al = a.lower()
        if no_wheel and al.endswith(".whl"):
            continue
        if module_name == "python" and ".msi" not in a:
            continue
        if "theme-" in al:
            al = al.split("theme-")
            al = al[0].replace("-", "_") + "theme-" + al[1]

        if al.startswith(module_name):
            return a
        if p and al.startswith(pp):
            return a
        if d and al.startswith(pd):
            return a
    return None


def is_package_installed(python_path, module_name, installed_packages=None):
    """
    not very accurate but it should speed up the process

    @param      python_path         python path
    @param      module_name         module name (import name)
    @param      installed_packages  list of installed packages (can be None)
    @return                         boolean
    """
    if isinstance(module_name, str  # unicode#
                  ):
        module_name = [module_name]
    modules = get_modules_version(python_path)
    for name in module_name:
        if name in modules:
            return True
        if installed_packages is not None and name in installed_packages:
            return True
        pymy = os.path.join(python_path, "lib", "site-packages", name)
        if os.path.exists(pymy):
            return True
        pymy += ".py"
        if os.path.exists(pymy):
            return True
        pymy = pymy.replace(".py", ".cp%d%d-win_amd64.pyd" %
                            (sys.version_info[0], sys.version_info[1]))
        if os.path.exists(pymy):
            return True
        mod = find_module_install(name)
        if mod:
            pymy = os.path.join(python_path, "lib", "site-packages", mod.mname)
            if os.path.exists(pymy):
                return True
    return False


def win_install_packages_other_python(python_path, package_folder, verbose=False, module_list=None, fLOG=print):
    """
    Install all packages for another Python distribution
    where package could be found in a folder

    @param      python_path     location of python
    @param      package_folder  location of the package (.tar.gz, .whl, .bz2, .tgz)
    @param      verbose         display more information
    @param      module_list     list of modules to install, if None, it tries to guess a good
                                order to install downloaded packages
    @param      fLOG            logging function
    @return                     operations ("pip", module) if installed, empty if already installed
    """
    files = os.listdir(package_folder)
    files = [_ for _ in files if os.path.splitext(_)[-1] in {".gz", ".zip", ".whl", ".bz2", ".tgz"} and
             _ not in {"Scite.zip", "scite.zip"} and
             not _.startswith("SQLiteSpy_")]

    # we need to order the package to install them in the right order
    # it speeds up the process and avoid using C++ compiler
    operations = []
    done = set()
    if module_list is None:
        full_list = small_set()
    else:
        full_list = module_list

    # existing list
    installed_packages = get_modules_version(python_path)

    for mod in full_list:
        a = _is_package_in_list(mod.name + "-", files)
        if a is None:
            continue
        if a not in done:
            mname = mod.mname if mod.mname is not None else mod.name
            if not is_package_installed(python_path, [mod.name, mname], installed_packages):
                full = os.path.join(package_folder, a)
                try:
                    op = win_install_package_other_python(
                        python_path, full, verbose=verbose, deps=mod.deps, fLOG=fLOG)
                except Exception as e:
                    mes = "failed to install {0}: {1}".format(mod.name, full)
                    raise Exception(mes) from e
                if len(op) > 0:
                    fLOG("[pymy] installed", mod.name, " with ", a)
                operations.extend(op)
            done.add(a)

    if verbose:
        fLOG("[pymy] *** install packages with unknown dependencies")
    for pack in files:
        if pack not in done:
            full = os.path.join(package_folder, pack)
            op = win_install_package_other_python(
                python_path, full, verbose=verbose, fLOG=fLOG)
            if len(op) > 0:
                fLOG("[pymy] installed", pack, " with ", full)
            else:
                fLOG("[pymy] skipped package", pack, " from ", full)
            operations.extend(op)

    return operations



"""
@file
@brief Various functions to install `MinGW <http://www.mingw.org/>`_.
"""
from __future__ import print_function
import sys
import os
import re

from .install_custom import download_page, download_file
from ..installhelper.install_cmd_helper import unzip_files


def install_chromedriver(dest_folder=".", fLOG=print, install=True, version=None):
    """
    Installs `chromedriver <https://sites.google.com/a/chromium.org/chromedriver/>`_.

    @param      dest_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      version         version to install (unused)
    @return                     zip file in a list or list of unzipped files

    This is required for `Selenium <https://selenium-python.readthedocs.io/>`_.
    """
    if version is None:
        url = "https://sites.google.com/a/chromium.org/chromedriver/"
        content = download_page(url)
        reg = re.compile(
            "(<a href=\\\"https://chromedriver."
            "storage.googleapis.com/"
            "index.html[?]path=([0-9]+([.][0-9]+){1,3})/)")
        f = reg.findall(content)
        if not f:
            raise Exception(
                "Unable to get the last version number for ChromeDriver, "
                "url='{}'".format(url))
        version = f[0][1]
    if sys.platform.startswith("win"):
        url = "http://chromedriver.storage.googleapis.com/{0}/chromedriver_win32.zip".format(
            version)
    elif sys.platform.startswith("mac"):
        url = "http://chromedriver.storage.googleapis.com/{0}/chromedriver_mac32.zip".format(
            version)
    else:
        url = "http://chromedriver.storage.googleapis.com/{0}/chromedriver_linux64.zip".format(
            version)
    name = url.split("/")[-1]

    outfile = os.path.join(dest_folder, name)
    if fLOG:
        fLOG("[pymy] chromedriver, download from ", url)
    download_file(url, outfile, fLOG=fLOG)

    if install:
        return unzip_files(outfile, whereTo=dest_folder, fLOG=fLOG)
    else:
        return [outfile]



"""
@brief      test log(time=2s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper import get_pypi_version


class TestPypiVersion(unittest.TestCase):

    def test_package_list(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        ver = get_pypi_version("virtualenv")
        if len(ver) == 0:
            raise Exception(str(ver))


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Caching functions
"""


def install_memoize(f):
    """
    cache a downloaded page
    """
    memo = {}

    def helper(x):
        if x not in memo:
            memo[x] = f(x)
        return memo[x]
    return helper


def install_memoize2(f):
    """
    cache a downloaded page
    """
    memo = {}

    def helper(x, is406=False):
        if x not in memo:
            memo[x] = f(x, is406)
        return memo[x]
    return helper



"""
@file
@brief Missing information about licenses
"""

missing_module_licenses = {
    "xgboost": "Apache-2",
    "actuariat_python": "MIT",
    "code_beatrix": "MIT",
    "ensae_teaching_cs": "MIT",
    "pymyinstall": "MIT",
    "pyensae": "MIT",
    "pyquickhelper": "MIT",
    "pymmails": "MIT",
    "pyrsslocal": "MIT",
    "pysqllike": "MIT",
    "spyre": "MIT",
    "dataspyre": "MIT",
    "ete": "GNU v3",
    "ete3": "GNU v3",
    "datrie": "LGPL v2.1",
    "glueviz": "BSD",
    "h5py": "BSD",
    "statsmodels": "BSD",
    "toolz": "BSD",
    "trackpy": "BSD - 3 clauses",
    "jupyter_core": "BSD",
    "jupyter_client": "BSD",
    "metakernel": "BSD",
    "scilab_kernel": "BSD",
    "pyexecjs": "MIT",
    "selenium": "Apache Software License",
    "alabaster": "BSD",
    "cloud_sptheme": "BSD",
    "colorama": "BSD",
    "guzzle_sphinx_theme": "BSD~MIT",
    "itcase_sphinx_theme": "no license",
    "pypiserver": "BSD",
    "solar_theme": "BSD~MIT",
    "sphinx_bootstrap_theme": "MIT",
    "sphinx_py3doc_enhanced_theme": "BSD",
    "sphinx_readable_theme": "MIT",
    "sphinxjp.themes.basicstrap": "MIT",
    "sphinxjp.themes.sphinxjp": "MIT",
    "wild_sphinx_theme": "BSD",
    "SQLAlchemy": "MIT",
    "graphviz": "MIT",
    "anyjson": "BSD",
    "brewer2mpl ": "MIT",
    "cached_property": "BSD",
    "celery": "BSD",
    "colorspacious": "MIT",
    "cgal_bindings": "Boost Software License 1.0",
    "Cython": "Apache Software License",
    "datashape": "BSD",
    "cubehelix": "~BSD - 2 clauses",
    "dbfread": "MIT",
    "dynd": "BSD",
    "fabric": "BSD",
    "feedparser": "~BSD - 2 clauses",
    "gmpy2": "LGPLv3+",
    "itsdangerous": "BSD",
    "kombu": "BSD",
    "lifelines": "MIT",
    "line_profiler": "BSD",
    "lxml": "BSD",
    "Mako": "MIT",
    "marisa_trie": "MIT",
    "mock": "BSD",
    "msgpack": "Apache Software License",
    "multimethods": "MIT",
    "multipledispatch": "BSD",
    "nodeenv": "BSD",
    "oauthlib": "BSD",
    "patsy": "BSD",
    "pbr": "Apache Software License",
    "pims": "~BSD",
    "pipdeptree": "MIT",
    "PyAudio": "MIT",
    "pycrypto": "Public Domain + patent for some algorithm",
    "planar": "BSD",
    "pygal_maps_world": "LGPLv3+",
    "pymesos": "BSD",
    "pystache": "MIT",
    "pytz": "MIT",
    "requests_oauthlib": "ISC",
    "rope_py3k": "GPL=OpenBSD",
    "smart_open": "Public Domain",
    "textblob": "MIT",
    "tzlocal": "CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
    "viscm": "MIT",
    "zs": "BSD",
    "neural-python": "MIT",
    "dask": "BSD",
    "NLopt": "LGPL",
    "ipython_genutils": "BSD",
    "memory_profiler": "BSD",
    "snakeviz": "BSD",
    "ansi2html": "GPLv3+",
    "brewer2mpl": "MIT",
    "dev": "Public Domain",
    "django-model-utils": "BSD",
    "django-userena": "BSD",
    "django-uuidfield": "~BSD",
    "easy-thumbnails": "BSD",
    "gunicorn": "MIT",
    "lockfile": "MIT",
    "gevent": "MIT",
    "lz4": "BSD",
    "opencv_python": "BSD",
    "ptyprocess": "ISC",
    "terminado": "BSD",
    "easy_thumbnails": "BSD",
    "llvmpy": "BSD",
    "pysterior": "MIT",
    "pymc3": "Apache-2",
    "tqdm": "MIT",
    "pandas-highcharts": "MIT",
    "holoviews": "BSD",
    "cyordereddict": "MIT",
    "future": "MIT",
    "python-gmaps": "BSD",
    "libpython": "Python",
    "pyowm": "MIT",
    "PyMySQL": "MIT",
    "osmapi": "GPLv3",
    "progressbar2": "BSD",
    "wordcloud": "MIT",
    "django-contrib-comments": "BSD",
    "filebrowser_safe": "~MIT",
    "grappelli_safe": "~MIT",
    "mezzanine": "BSD",
    "affine": "BSD",
    "blocks": "GPL",
    "blz": "BSD",
    "chest": "BSD",
    "configobj": "BSD",
    "distributed": "BSD",
    "gatspy": "BSD",
    "guidata": "CeCILL v2",
    "guiqwt": "CECILL",
    "heapdict": "BSD",
    "httpretty": "MIT",
    "jaraco.structures": "MIT",
    "jdcal": "BSD",
    "jmespath": "MIT",
    "keyring": "MIT, Python",
    "locket": "BSD",
    "onedrive-sdk-python": "MIT",
    "pandocfilters": "BSD",
    "passlib": "BSD",
    "picklable-itertools": "MIT",
    "plac": "BSD",
    "pycryptodomex": "BSD",
    "PyOpenGL_accelerate": "BSD",
    "pyOpenSSL": "Apache 2.0",
    "pyRFXtrx": "LGPLv3+",
    "pysmi": "BSD",
    "pythonnet": "MIT",
    "pythonqwt": "MIT",
    "queuelib": "BSD",
    "semantic_version": "BSD",
    "slicerator": "BSD",
    "sqlite_bro": "MIT",
    "supersmoother": "BSD 3-clause",
    "tblib": "BSD",
    "unidecode": "GPLv2+",
    "vincenty": "none",
    "user-agent": "MIT",
    "wget": "none",
    "xxhash": "BSD",
    "zope.exceptions": "Zope Public License",
    "w3lib": "BSD",
    "astropy": "BSD",
    "backports_abc": "Python Software Foundation License",
    "boto3": "Apache Software License ",
    "botocore": "Apache Software License ",
    "scoop": "GNU Library or Lesser General Public License (LGPL)",
    "hikvision": "MIT",
    "PyMata": "GPLv3+",
    "python-mpd2": "LGPL",
    "python-nest": "Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License",
    "temperusb": "GNU GENERAL PUBLIC LICENSE",
    "sphinx-corlab-theme": "LGPLv3+",
    "sphinx-docs-theme": "MIT",
    "sphinxjp.themes.gopher": "MIT",
    "sphinxjp.themes.htmlslide": "MIT",
    "zerovm-sphinx-theme": "Apache License 2.0",
    "xarray": "Apache Software License",
    "typing": "PSF",
    "sklearn_pandas": "~MIT",
    "skll": "BSD",
    "abcpmc": "GPLv3+",
    "ad3": "GPLv3",
    "amqp": "LGPL",
    "ansiconv": "MIT",
    "apscheduler": "MIT",
    "autopy3": "MIT",
    "azure-batch-apps": "MIT",
    "azureml": "MIT",
    "bigfloat": "LGPLv3",
    "billiard": "BSD",
    "biopython": "~MIT",
    "bleach": "Apache",
    "blist": "BSD",
    "blosc": "MIT",
    "bqplot": "Apache",
    "btrees": "Zope Public License",
    "bz2file": "Apache",
    "Cartopy": "LGPLv3+",
    "chalmers": "MIT",
    "CherryPy": "BSD",
    "cobble": "BSD",
    "comtypes": "MIT",
    "contextlib2": "Python",
    "cssselect": "BSD",
    "cubes": "MIT",
    "cuda4py": "BSD",
    "CVXcanon": "?",
    "cvxpy": "GNU",
    "cymem": "MIT",
    "datashader": "BSD",
    "db.py": "BSD",
    "deap": "LGPL",
    "Django": "BSD",
    "django-audiotracks": "MIT",
    "django-celery": "BSD",
    "django-configurations": "BSD",
    "django-environ": "MIT",
    "django-guardian": "BSD",
    "django-storages": "BSD",
    "dlib": "boost",
    "dnspython": "~MIT",
    "docrepr": "Modified BSD",
    "entrypoints": "MIT",
    "envoy": "~MIT",
    "et_xmlfile": "MIT",
    "Flask-Login": "MIT",
    "Flask-SQLAlchemy": "BSD",
    "GDAL": "MIT",
    "gensim": "LGPLv2+",
    "geoplotlib": "MIT",
    "geopy": "MIT",
    "google-api-python-client": "Apache",
    "googlemaps": "Apache",
    "grab": "MIT",
    "greenlet": "MIT",
    "grequests": "BSD",
    "grin": "BSD",
    "HDDM": "BSD",
    "heatmap": "MIT",
    "Hebel": "GPLv2",
    "hmmlearn": "MIT + Copyright",
    "html2text": "GPL",
    "html5lib": "MIT",
    "httpie": "BSD",
    "imageio": "BSD",
    "imbox": "MIT",
    "invoke": "BSD",
    "ipyparallel": "BSD",
    "jedi": "MIT",
    "jieba": "MIT",
    "julia": "MIT",
    "jupytalk": "MIT",
    "Keras": "MIT",
    "Kivy": "MIT",
    "kivy-garden": "MIT",
    "luigi": "Apache 2.0",
    "ldap3": "LGPLv3",
    "lazy-object-proxy": "BSD",
    "AnyQt": "GPLv3",
    "Automat": "MIT",
    "azure-datalake-store": "MIT",
    "azure-keyvault": "MIT",
    "azure-mgmt-containerregistry": "MIT",
    "azure-mgmt-datalake-analytics": "MIT",
    "azure-mgmt-datalake-nspkg": "MIT",
    "azure-mgmt-datalake-store": "MIT",
    "azure-mgmt-devtestlabs": "MIT",
    "azure-mgmt-dns": "MIT",
    "azure-mgmt-documentdb": "MIT",
    "azure-mgmt-iothub": "MIT",
    "azure-mgmt-keyvault": "MIT",
    "azure-mgmt-monitor": "MIT",
    "azure-mgmt-rdbms": "MIT",
    "azure-mgmt-sql": "MIT",
    "azure-mgmt-trafficmanager": "MIT",
    "azure-servicefabric": "MIT",
    "backports.functools_lru_cache": "MIT",
    "bcrypt": "Apache License 2.0",
    "blockdiag": "Apache License 2.0",
    "Bottlechest": "Simplified BSD",
    "CacheControl": "Apache Software License",
    "category_encoders": "BSD",
    "Cheroot": "BSD",
    "colorcet": "Creative Commons Attribution 4.0 International Public License (CC-BY)",
    "colorlover": "MIT",
    "defusedxml": "Python Software Foundation License Version 2",
    "docx2txt": "MIT",
    "dominate": "GNU Lesser General Public License (LGPL)",
    "drawtree": "Apache License 2.0",
    "edward": "MIT",
    "elasticsearch": "Apache License 2.0",
    "ensae_projects": "MIT",
    "Fabric3": "BSD",
    "fastparquet": "Apache License 2.0",
    "fasttext": "BSD 3-Clause License",
    "fbprophet": "BSD",
    "filelock": "Unlicensed",
    "ftfy": "MIT",
    "funcparserlib": "MIT",
    "gitdb": "BSD",
    "GitPython": "BSD",
    "glue-core": "BSD",
    "glue-vispy-viewers": "As Is",
    "gvar": "GPLv3+",
    "h2": "MIT",
    "holopy": "GNU GENERAL PUBLIC LICENSE",
    "hpack": "MIT",
    "hyper": "MIT",
    "hyperframe": "MIT",
    "hyperlink": "MIT",
    "hyperspy": "GNU General Public License v3 (GPLv3)",
    "images2gif": "Unlicensed",
    "ipaddress": "Python Software Foundation License",
    "JPype1": "Apache Software License",
    "jsonpickle": "BSD",
    "jupyter_contrib_nbextensions ": "BSD",
    "jupyter_sphinx": "BSD",
    "jyquickhelper": "MIT",
    "kabuki": "As Is",
    "libLAS": "BSD",
    "lmfit": "BSD",
    "lsqfit": "GPLv3+",
    "lru-dict": "MIT",
    "mammoth": "BSD",
    "mbstrdecoder": "MIT",
    "mdn-sphinx-theme": "Mozilla Public License 2.0 (MPL 2.0)",
    "missingno": "As Is",
    "mkl-service": "Anaconda End User License Agreement",
    "mlstatpy": "MIT",
    "monotonic": "Apache",
    "monty": "MIT",
    "msgpack-numpy": "BSD",
    "nbdime": "BSD",
    "nose-parameterized": "BSD",
    "octave_kernel": "BSD",
    "odfpy": "GNU General Public License v.2 + Apache License v.2",
    "Orange3": "GPLv3+",
    "Orange3-Associate": "GPLv3+",
    "Orange3-ImageAnalytics": "GPLv3+",
    "Orange3-Network": "GPLv3+",
    "Orange3-Text": "GPLv3+",
    "palettable": "MIT",
    "path.py": "MIT",
    "pathvalidate": "MIT",
    "pdfminer3k": "MIT",
    "pdfrw": "MIT",
    "PIMS": "As Is",
    "Pint": "BSD",
    "pybars3": "GNU Library or Lesser General Public License (LGPL)",
    "pyclustering": "GNU General Public License v3 (GPLv3)",
    "pyemd": "MIT",
    "pyflux": "As Is",
    "pygal_maps_ch": "GNU LGPL v3+",
    "pygal_maps_fr": "GNU LGPL v3+",
    "pygal_sphinx_directives": "GNU LGPL v3+",
    "pyinstrument": "BSD",
    "pymatgen": "MIT",
    "pyopencl": "MIT",
    "PyOpenGL": "BSD",
    "pyPdf": "BSD",
    "PyPDF2": "BSD",
    "pypmc": "GPLv2",
    "pypng": "MIT",
    "pystan": "GNU General Public License v3 (GPLv3)",
    "PySide": "GNU Library or Lesser General Public License (LGPL)",
    "pytablereader": "MIT",
    "python-mimeparse": "MIT",
    "python3-linkedin": "MIT",
    "python-Levenshtein": "GNU General Public License v2 or later (GPLv2+)",
    "pythreejs": "BSD",
    "PyX": "GNU General Public License (GPL)",
    "qutip": "BSD",
    "recommonmark": "MIT",
    "regex": "Python Software Foundation License",
    "Rtree": "GNU Library or Lesser General Public License (LGPL)",
    "sacred": "MIT",
    "scikit-fusion": "GPLv3",
    "service_identity": "MIT",
    "setproctitle": "BSD",
    "setuptools-git": "BSD",
    "sdepy": "BSD",
    "simhash": "MIT",
    "SimpleSQLite": "MIT",
    "SIP": "GPL v2 or GPL v3 or BSD",
    "smmap": "BSD",
    "sounddevice": "MIT",
    "spglib": "BSD",
    "sphinx_theme_pd": "MIT",
    "sphinxcontrib-blockdiag": "BSD",
    "splinter": "As Is",
    "TA-Lib": "BSD",
    "teachpyx": "MIT",
    "tempora": "MIT",
    "tensorflow": "Apache Software License",
    "termcolor": "MIT",
    "testpath": "MIT",
    "tkinterquickhelper": "MIT",
    "toml": "MIT",
    "toyplot": "BSD",
    "traits": "BSD",
    "triangle": "GNU LGPL",
    "typepy": "MIT",
    "update_checker": "Simplified BSD License",
    "validate_email": "LGPL",
    "webcolors": "BSD",
    "wikipedia": "MIT",
    "win_unicode_console": "MIT",
    "xlwings": "BSD",
}



"""
@file
@brief Various functions to install `MinGW <http://www.mingw.org/>`_.
"""
from __future__ import print_function
import sys
import os

from .install_custom import download_file
from ..installhelper.install_cmd_helper import run_cmd


def install_vs(dest_folder=".", fLOG=print, install=True, version=None):
    """
    install `Visual Studio Express <https://www.visualstudio.com/en-us/products/visual-studio-express-vs.aspx>`_ (only on Windows)

    @param      dest_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      version         version to download (unused)
    @return                     temporary file

    @todo check it is installed
    """
    if version is not None:
        raise ValueError("cannot specify a version")
    if not sys.platform.startswith("win"):
        raise NotImplementedError(
            "Visual Studio can only be installed on Windows at the moment")

    name = "vs_community.exe"
    newurl = "https://go.microsoft.com/fwlink/?LinkId=532606&clcid=0x40c"
    outfile = os.path.join(dest_folder, name)
    fLOG("[pymy] Visual Studio, download from ", newurl)
    local = download_file(newurl, outfile)

    if install:
        run_cmd("msiexec /i " + local, fLOG=fLOG, wait=True)
    return local



"""
@file
@brief Various functions to install `Putty <http://www.putty.org/>`_.
"""
from __future__ import print_function
import sys
import re
import os

from .install_custom import download_page, download_file

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


def IsPuttyInstalled(dest_folder):
    """
    check if Scite was already installed

    @param      dest_folder     where it was installed
    @return                     boolean
    """
    if sys.platform.startswith("win"):
        file = os.path.join(dest_folder, "putty.exe")
        return os.path.exists(file)
    else:
        raise NotImplementedError("not available on platform " + sys.platform)


def install_putty(dest_folder=".", fLOG=print, install=True, version=None):
    """
    install `Putty <http://www.putty.org/>`_ (only on Windows)

    @param      dest_folder     where to download putty
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      version         version to install (unused)
    @return                     temporary file
    """
    if version is not None:
        raise ValueError("cannot specify a version")
    if IsPuttyInstalled(dest_folder):
        return os.path.join(
            os.path.abspath(dest_folder), "putty.exe")

    if not sys.platform.startswith("win"):
        raise NotImplementedError(
            "SciTE can only be installed on Windows at the moment")

    url = "http://www.chiark.greenend.org.uk/~sgtatham/putty/latest.html"
    page = download_page(url)

    reg = re.compile(
        "<a href=\\\"(.*latest/w64/putty.exe)\\\"><code>putty.exe</code></a>")
    find = reg.findall(page)
    if len(find) != 1:
        mes = "unable to find the file to download at {0}\nfound: {1}\npattern: {2}\nOUT:\n{3}".format(
            url, len(find), reg.pattern, "\n".join(find))
        raise Exception(mes)

    # should be something like http://www.scintilla.org/wscite356.zip
    newurl = find[0]
    outfile = os.path.join(dest_folder, "putty.exe")
    if not os.path.exists(outfile):
        try:
            download_file(newurl, outfile)
        except Exception as e:
            raise Exception("unable to download\n{0}\nto{1}".format(
                newurl, outfile)) from e

        if not os.path.exists(outfile):
            raise FileNotFoundError(outfile)

    return outfile



# -*- coding: utf-8 -*-
import sys
import os
import alabaster
from pyquickhelper.helpgen.default_conf import set_sphinx_variables

sys.path.insert(0, os.path.abspath(os.path.join(os.path.split(__file__)[0])))

set_sphinx_variables(__file__, "pymyinstall", "Xavier Dupré", 2022,
                     "alabaster", alabaster.get_path(),
                     locals(), add_extensions=None,
                     extlinks=dict(issue=('https://github.com/sdpython/pymyinstall/issues/%s', 'issue')))

blog_root = "http://www.xavierdupre.fr/app/pymyinstall/helpsphinx/"


epkg_dictionary.update({
    'Azure': 'https://github.com/Azure/azure-sdk-for-python',
    'avconv': 'http://debian-facile.org/doc:media:avconv',
    'boost': "http://www.boost.org/",
    'boost_python': "http://www.boost.org/doc/libs/release/libs/python/",
    'cartopy': 'https://scitools.org.uk/cartopy/docs/latest/',
    'clang': 'https://clang.llvm.org/get_started.html',
    'cmake': 'https://cmake.org/',
    'cvxopt': "http://cvxopt.org/",
    'cvxpy': "http://www.cvxpy.org/en/latest/index.html",
    'Debian 9': 'https://www.debian.org/News/2017/20170617',
    'dlib': "http://dlib.net/",
    'dotnet': 'https://www.microsoft.com/net',
    'ffmpeg': 'https://www.ffmpeg.org/',
    'java': 'https://www.java.com/',
    'imageio': 'https://imageio.github.io/',
    'keyring': 'https://github.com/jaraco/keyring',
    'lighttpd': 'https://www.lighttpd.net/',
    'line_profiler': 'https://github.com/rkern/line_profiler',
    'minepy': 'https://github.com/minepy/minepy',
    'Miniconda': 'https://docs.conda.io/en/latest/miniconda.html',
    'MKL': 'https://software.intel.com/en-us/mkl',
    'ML.net': 'https://www.microsoft.com/net/learn/apps/machine-learning-and-ai/ml-dotnet',
    'mono': 'https://www.mono-project.com/',
    'nginx': 'https://www.nginx.com/',
    'nodejs': 'https://nodejs.org/en/',
    'onnx': 'https://onnx.ai/',
    'pip': 'https://pip.pypa.io/en/stable/',
    'protobuf': 'https://github.com/protocolbuffers/protobuf',
    'pyaudio': 'https://people.csail.mit.edu/hubert/pyaudio/docs/',
    'pypi': 'https://pypi.org/',
    'pycuda': 'https://mathema.tician.de/software/pycuda/',
    'pycurl': 'http://pycurl.io/',
    'pygame': 'https://www.pygame.org/',
    'pyopencl': 'https://documen.tician.de/pyopencl/',
    'pyproj': 'https://github.com/pyproj4/pyproj',
    'pythonnet': 'https://github.com/pythonnet/pythonnet',
    'spacy': 'https://spacy.io/',
    'SQLiteSpy': "https://www.yunqa.de/delphi/products/sqlitespy/index",
    'statsmodels': "http://www.statsmodels.org/stable/index.html",
    'tables': 'https://www.pytables.org/usersguide/tutorials.html',
    'TensorFlow': 'https://www.tensorflow.org/',
    'ufw': 'https://doc.ubuntu-fr.org/ufw',
    'unrar': 'https://en.wikipedia.org/wiki/Unrar',
    'virtualenv': 'https://virtualenv.pypa.io/en/stable/',
})



"""
@brief      test log(time=1s)
"""
import unittest
import warnings
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.module_install import ModuleInstall
from pymyinstall.installhelper.module_install_exceptions import MissingVersionOnPyPiException


class TestPyPiModule (unittest.TestCase):

    def test_pipy_bug_flask(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mod = ModuleInstall(
            'flask-sqlalchemy', 'pip', mname='flask.ext.sqlalchemy')
        vers = mod.get_pypi_version()
        assert vers is not None

    def test_pipy_bug_ipython(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mod = ModuleInstall(
            'ipython', 'wheel', mname='IPython')
        try:
            vers = mod.get_pypi_version()
        except MissingVersionOnPyPiException:
            vers = "no release version"
            warnings.warn("IPython is is unstable, transitionning to jupyter")
        assert vers is not None


if __name__ == "__main__":
    unittest.main()



#!python
"""
script which goes through all installed modules
"""
from __future__ import print_function
import sys
import os
import argparse


def get_parser():
    """
    defines the way to parse the script ``pymy_status``
    """
    typstr = str  # unicode#
    parser = argparse.ArgumentParser(
        description='information about installed modules')
    parser.add_argument(
        '-o',
        '--out',
        default="python_module.xlsx",
        type=typstr,
        help='output the results into a file (required pandas)')
    parser.add_argument(
        '-p',
        '--pypi',
        default=True,
        type=bool,
        help='checks the version on PyPi')
    parser.add_argument(
        'module',
        nargs="*",
        default="all",
        help='update only the list of modules included in this list or all modules if not specified or equal to all')
    parser.add_argument(
        '--set',
        default="-",
        type=typstr,
        help='set of module to install, see documentation of function get_name_set to get a comprehensive list, ' +
             'this option is ignored if a module is specified on the command line')
    return parser


def do_main(list_module=None, outfile="python_module.xlsx", pypi=True):
    """
    Calls function @see fn update_all but is meant to be added to scripts folder.

    @param      list_module     list of modules to update or None for all
    @param      outfile         output the results into a flat file or an excel file (required pandas)
    @param      pypi            check version on PyPi
    """
    try:
        from pymyinstall.installhelper import get_installed_modules
    except ImportError:
        pfolder = os.path.normpath(os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "..", ".."))
        sys.path.append(pfolder)
        if "pymyinstall" in sys.modules:
            del sys.modules["pymyinstall"]
        if "pymyinstall.installhelper" in sys.modules:
            del sys.modules["pymyinstall.installhelper"]
        from pymyinstall.installhelper import get_installed_modules
    if list_module is not None and len(list_module) == 0:
        list_module = None
    if outfile:
        import pandas
    res = get_installed_modules(short_list=list_module, fLOG=print, pypi=True)
    if outfile:
        df = pandas.DataFrame(res)
        cols = list(df.columns)
        cols.sort()
        df = df[cols]
        if ".xls" in os.path.split(outfile)[-1]:
            print("Saving into:", outfile)
            df.to_excel(outfile, index=False)
        else:
            print("Saving into (sep=tab, encoding=utf-8):", outfile)
            df.to_csv(outfile, seo="\t", encoding="utf-8", index=False)


def main():
    """
    calls function @see fn update_all but is meant to be added to scripts folder,
    parse command line arguments
    """
    parser = get_parser()
    try:
        res = parser.parse_args()
    except SystemExit:
        print(parser.format_usage())
        res = None

    if res is not None:
        list_module = None if res.module in [
            "all", "", "-", None, []] else res.module
        if list_module is None and res.set is not None and len(res.set) > 0 and res.set != "-":
            list_module = res.set
        do_main(list_module=list_module, outfile=res.out, pypi=res.pypi)


if __name__ == "__main__":
    main()



# coding: utf-8
"""
@file
@brief Contains a list of web site with some useful tools.
"""
import webbrowser

_tools = {
    "7z": "http://www.7-zip.org/download.html",
    "miktex": "http://miktex.org/download",
    "texniccenter": "http://www.texniccenter.org/download/",
    "python": "https://www.python.org/downloads/release/python-341/",
    "mingw": "http://sourceforge.net/projects/mingw/files/Installer/",
    "VS": "http://www.microsoft.com/download/details.aspx?id=40787",
    "cygwin": "https://cygwin.com/install.html",
    "chrome": "http://www.google.com/chrome/browser/",
    "firefox": "http://www.mozilla.org/firefox/new/",
    "graphviz": "http://www.graphviz.org/",
    "doxygen": "http://www.stack.nl/~dimitri/doxygen/download.html",
    "nodejs": "http://nodejs.org/download/",
    # "innosetup":"http://www.jrsoftware.org/isdl.php",
}


def get_install_list():
    """
    returns the list of tools a developper might need
    """
    global _tools  # pylint: disable=W0602
    return list(_tools.keys())


def open_tool_on_browser(tool=None):
    """
    open a page on browser for a specific tool

    @param      tool        tool name
    """
    global _tools  # pylint: disable=W0602
    if isinstance(tool, str):
        tool = [tool]
    elif tool is None:
        tool = get_install_list()
    for t in tool:
        webbrowser.open_new_tab(_tools[t])


if __name__ == "__main__":
    open_tool_on_browser()



"""
@brief      test log(time=3s)
"""
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.setuphelper import create_empty_folder_setup


class TestOnlyMain (unittest.TestCase):

    def test_empty_setup(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        temp = get_temp_folder(__file__, "temp_empty_setup")

        cr = create_empty_folder_setup(temp, "CVXcanon",
                                       description="module created with compiled files issues from the installation of CVXcanon")
        assert len(cr) >= 2
        for o in cr:
            if not os.path.exists(o):
                raise FileNotFoundError(o)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=45s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.link_shortcuts import suffix


class Testinks (unittest.TestCase):

    def test_links(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        suf = suffix()
        fLOG("*", suf)
        self.assertTrue("." in suf)
        self.assertEqual(len(suf.split(".")), 3)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=10s)
"""
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import check_pep8, ExtTestCase


class TestCodeStyle(ExtTestCase):

    def test_style_src(self):
        thi = os.path.abspath(os.path.dirname(__file__))
        src_ = os.path.normpath(os.path.join(thi, "..", "..", "src"))
        check_pep8(src_, fLOG=fLOG,
                   pylint_ignore=('C0103', 'C1801', 'R0201', 'R1705', 'W0108', 'W0613',
                                  'R1703', 'C0111', 'C0201', 'R1702',
                                  'W0703', 'W0622', 'C0412', 'R0912',
                                  'W0603', 'W0212', 'W0123', 'W0201',
                                  'C0200', 'W0122', 'C0302', 'W0102',
                                  'W0621', 'R1710', 'R0401', 'R1714',
                                  'R0915', 'W0107', 'W503', 'C0415',
                                  'W0707', 'R1725', 'R1732', 'C0207',
                                  'W1514', 'C0209'),
                   skip=["__init__.py:1: R0401",
                         "win_installer",
                         "Redefining built-in ",
                         "is assigned to nothing",
                         "Disallow trailing comma tuple",
                         "The if statement can be replaced with",
                         "Unable to import 'src.pymyinstall.packaged.get_pip'",
                         "Redefining name 'is_installed' from outer",
                         "module_install_version.py:17: E0401",
                         "module_install_page_wheel.py:15: E0401",
                         "module_install_page_wheel.py:14: E0401",
                         "install_custom.py:13: E0401",
                         "install_custom.py:14: E0401",
                         "install_custom.py:15: E0401",
                         "install_custom_python.py:220",
                         "module_install_version.py:16: E0401",
                         "module_install_page_wheel.py:13: E0401",
                         "module_install_page_wheel.py:13: E0401",
                         "module_install.py:29: E0401",
                         "module_install.py:28: E0401",
                         "module_install.py:27: E0401",
                         "get_pip.py",
                         "pymy_status.py:53: W0612",
                         "module_install.py:26: E0401",
                         "ipython_helper.py:24: E1101",
                         "Unable to import 'selenium.webdriver'",
                         "Unable to import 'Queue'",
                         "Unable to import 'src.pymyinstall.installhelper.status_helper'",
                         "[W503] line break before binary operator",
                         "install_custom_python.py:164",
                         "ipython_helper.py:21: E1101",
                         ])

    def test_style_test(self):
        thi = os.path.abspath(os.path.dirname(__file__))
        test = os.path.normpath(os.path.join(thi, "..", ))
        check_pep8(test, fLOG=fLOG, neg_pattern="temp_.*",
                   pylint_ignore=('C0103', 'C1801', 'R0201', 'R1705', 'W0108', 'W0613',
                                  'C0111', 'W0703', 'W0622', 'W0621', 'C0412', 'C0411',
                                  'R0915', 'W0107', 'W503', 'C0415', 'W0707', 'R1725',
                                  'R1732', 'C0207', 'W1514', 'C0209'),
                   skip=["test_is_installed.py:41: E1101",
                         "Unable to import 'selenium.webdriver'",
                         "Unable to import 'Queue'",
                         ])


if __name__ == "__main__":
    unittest.main()



#!python
"""
script which updates all modules,
works on Linux and Windows if the module is
included the list of modules handled by this module
"""
from __future__ import print_function
import sys
import os
import argparse


def get_parser():
    """
    defines the way to parse the script ``pymy_update``
    """
    typstr = str  # unicode#
    parser = argparse.ArgumentParser(
        description='update modules, consider wheels when the module includes C++ files')
    parser.add_argument(
        '-s',
        '--skip',
        default="ete,dataspyre,pycuda,cubehelix",
        type=typstr,
        help='list of modules to skip (not to be updated) separated by a comma')
    parser.add_argument(
        '-f',
        '--folder',
        type=typstr,
        default="build/update_modules",
        help='folder where modules will be downloaded')
    parser.add_argument(
        '--set',
        default="-",
        type=typstr,
        help='set of module to install, see documentation of function get_name_set to get a comprehensive list, ' +
             'this option is ignored if a module is specified on the command line')
    parser.add_argument(
        '--source',
        default="",
        type=typstr,
        help='overwrite the source of the wheels')
    parser.add_argument(
        '--schedule',
        action='store_true',
        help='do not update the modules, returned the list scheduled to be updated')
    parser.add_argument(
        'module',
        nargs="*",
        default="all",
        help='update only the list of modules included in this list or all modules if not specified or equal to all')
    return parser


def do_main(temp_folder="build/update_modules",
            skip_module="ete dataspyre pycuda cubehelix".split(),
            list_module=None, schedule_only=False, source=None):
    """
    calls function @see fn update_all but is meant to be added to scripts folder

    @param      temp_folder     folder where modules will be downloaded
    @param      skip_module     skip the module on this list
    @param      list_module     list of modules to update or None for all
    @param      schedule_only   if True, the function returns the list of modules scheduled to be installed
    @param      source          overwrite the source of the wheels
    """
    try:
        from pymyinstall import is_travis_or_appveyor
    except ImportError:
        def is_travis_or_appveyor():
            if "travis" in sys.executable:
                return "travis"
            if os.environ.get("USERNAME", os.environ.get("USER", None)) == "appveyor":
                return "appveyor"
            return None
    if is_travis_or_appveyor() and source is None:
        source = "2"

    if not os.path.exists(temp_folder):
        os.makedirs(temp_folder)
    try:
        from pymyinstall.packaged import update_all
    except ImportError:
        pfolder = os.path.normpath(os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "..", ".."))
        sys.path.append(pfolder)
        if "pymyinstall" in sys.modules:
            del sys.modules["pymyinstall"]
        if "pymyinstall.packaged" in sys.modules:
            del sys.modules["pymyinstall.packaged"]
        from pymyinstall.packaged import update_all
    res = update_all(temp_folder=temp_folder, verbose=True,
                     skip_module=skip_module, list_module=list_module,
                     schedule_only=schedule_only, source=source)
    if schedule_only:
        print("SCHEDULED")
        for r in res:
            print(r)


def main():
    """
    calls function @see fn update_all but is meant to be added to scripts folder,
    parse command line arguments
    """
    parser = get_parser()
    try:
        res = parser.parse_args()
    except SystemExit:
        print(parser.format_usage())
        res = None

    if res is not None:
        skip_module = None if res.skip in [
            "", "-", None, []] else res.skip.split(",")
        list_module = None if res.module in [
            "all", "", "-", None, []] else res.module
        if list_module is None and res.set is not None and len(res.set) > 0 and res.set != "-":
            list_module = res.set
        do_main(temp_folder=res.folder, skip_module=skip_module,
                list_module=list_module, schedule_only=res.schedule,
                source=res.source if res.source else None)


if __name__ == "__main__":
    main()



"""
@brief      test log(time=2s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder


class TestMakeExe(unittest.TestCase):

    def test_make_exe(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        temp = get_temp_folder(__file__, "temp_make_exe")

        from pip._vendor.distlib.scripts import ScriptMaker
        maker = ScriptMaker(os.path.join(temp, "..", "..", "..", "src"), temp)
        maker.make('run_update_all = pymyinstall:update_all')
        if sys.platform.startswith("win"):
            exe = os.path.join(temp, "run_update_all.exe")
            assert os.path.exists(exe)


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Various functions to install `Jenkins <http://jenkins-ci.org/>`_.

.. versionadded:: 1.1
"""
from __future__ import print_function
import sys
import os

from .install_custom import download_file


def install_jenkins(dest_folder=".", fLOG=print, install=True, version=None):
    """
    install `Jenkins <http://jenkins-ci.org/>`_ (only on Windows)

    @param      dest_folder         where to download the setup
    @param      fLOG                logging function
    @param      install             install (otherwise only download)
    @param      version             version to install (unused)
    @return                         temporary file

    .. versionadded:: 1.1
    """
    if version is not None:
        raise ValueError("cannot specify a version")

    if not sys.platform.startswith("win"):
        raise NotImplementedError(
            "SciTE can only be installed on Windows at the moment")

    url = "http://mirrors.jenkins.io/war/latest/jenkins.war"

    outfile = os.path.join(dest_folder, "jenkins.war")
    if not os.path.exists(outfile):
        download_file(url, outfile)

    if install:
        raise NotImplementedError("Does not install jenkins.war")
    return outfile



"""
@file
@brief Various functions to install `TDM-GCC <http://tdm-gcc.tdragon.net/>`_.
"""
from __future__ import print_function
import sys
import os
import re

from .install_custom import download_page, download_from_sourceforge


def install_tdm_gcc(dest_folder=".", fLOG=print, install=False, version=None):
    """
    install `TDM-GCC <http://tdm-gcc.tdragon.net/>`_ (only on Windows)

    @param      dest_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      version         version to download (unused)
    @return                     downloaded file

    It should
    ::

        mingw-get install binutils gcc g++ mingw32 fortran gdb mingw32 mingw w32api g77==3.4.5

    """
    if version is not None:
        raise ValueError("cannot specify a version")
    if not sys.platform.startswith("win"):
        raise NotImplementedError(
            "TDM-GCC can only be installed on Windows at the moment")

    url = "http://tdm-gcc.tdragon.net/download"
    page = download_page(url, is406=True)

    reg = re.compile(
        "<a href=\\\"http://sourceforge.net/projects/tdm-gcc/files/TDM-GCC[%]20Installer/(tdm64-.*?[.]exe)/download\\\"")
    find = reg.findall(page)
    if len(find) == 0:
        raise Exception("unable to find the file to download at " +
                        url + "\nfound: " + str(len(find)) + "\n" + "\n".join(find))
    name = find[0]

    newurl = "http://sourceforge.net/projects/tdm-gcc/files/TDM-GCC%20Installer/{0}/download?use_mirror=autoselect".format(
        name)
    outfile = os.path.join(dest_folder, name)
    fLOG("[pymy] tdm-gcc, download from ", newurl)
    file = download_from_sourceforge(
        newurl,
        outfile,
        fLOG=fLOG,
        temp_folder=dest_folder)

    if install:
        raise NotImplementedError()
    return file



"""
@brief      test log(time=131s)
"""
import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installcustom import install_graphviz


class TestGraphviz(unittest.TestCase):

    def test_install_graphviz(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if sys.version_info[0] == 2:
            return

        fold = os.path.abspath(os.path.split(__file__)[0])
        temp = os.path.join(fold, "temp_graphviz")
        if not os.path.exists(temp):
            os.mkdir(temp)
        for _ in os.listdir(temp):
            f = os.path.join(temp, _)
            if os.path.isfile(f):
                os.remove(f)

        if sys.platform.startswith("win"):
            r = install_graphviz(temp, fLOG=fLOG, install=False, source="2")
            exe = os.path.abspath(r)
            self.assertTrue(os.path.exists(exe))

    def test_install_graphviz_zip(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if sys.version_info[0] == 2:
            return

        temp = get_temp_folder(__file__, "temp_graphviz_zip")

        if sys.platform.startswith("win"):
            r = install_graphviz(temp, fLOG=fLOG, install=True, source="zip")
            exe = os.path.abspath(r)
            self.assertTrue(os.path.exists(exe))
            exe = os.path.join(temp, "Graphviz", "bin", "dot.exe")
            self.assertTrue(os.path.exists(exe))


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.module_install import ModuleInstall
from pymyinstall.installhelper.module_install_version import get_module_version


class TestLONGInstallModule (unittest.TestCase):

    def test_pypi_version(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mod = ModuleInstall("sphinx", "pip")
        vers = mod.get_pypi_version()
        assert vers >= "1.3.1"

        mod = ModuleInstall("scikit-learn", "wheel", mname="sklearn")
        fLOG(mod)
        vers = mod.get_pypi_version()
        assert vers >= "0.16.1"
        update = mod.has_update()
        fLOG("scikit-learn", update)

        mod = ModuleInstall("pandas", "wheel")
        vers = mod.get_pypi_version()
        if vers is None or vers < "0.16.1":
            raise Exception("{0}: {1}".format(mod.name, vers))

        mod = ModuleInstall("openpyxl", "pip", version="2.3.5")
        fLOG(mod)
        vers = mod.get_pypi_version()
        if vers is None or vers < "2.3.5":
            raise Exception("{0}: {1}".format(mod.name, vers))

        update = mod.has_update()
        if update:
            vers = mod.get_pypi_numeric_version()
            fLOG(vers)
        fLOG(update)

    def test_installed_version(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mod = ModuleInstall("jinja2", "pip")
        res = mod.is_installed_version()
        if not res:
            fLOG(mod)
            fLOG(mod.get_installed_version())
            for k, v in get_module_version(None).items():
                if k[0] in ("j", "J"):
                    fLOG("+++", k, v)
            assert False

        mod = ModuleInstall("pandas", "wheel")
        res = mod.is_installed_version()
        if res is None:
            raise AssertionError(
                "Issue with module (res empty) %r." % mod)
        fLOG("****", mod.get_installed_version(), mod.get_pypi_version())
        if (mod.get_installed_version() != mod.get_pypi_version() and
                mod.get_pypi_version() not in str(mod.get_installed_version())):
            if not mod.has_update():
                raise AssertionError(
                    "Issue with module %r - (%r != %r)." % (
                        mod, mod.get_installed_version(),
                        mod.get_pypi_version()))


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=3s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.win_installer.win_ipy_kernels import add_kernel_jupyter, python_kernel


class TestKernels(unittest.TestCase):

    def test_kernels(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        temp = get_temp_folder(__file__, "temp_kernels")
        ker = add_kernel_jupyter(
            python_kernel, temp, temp, os.path.dirname(sys.executable), suffix="ZZZ")
        with open(ker, "r", encoding="utf8") as f:
            content = f.read()
        fLOG(content)
        assert '"IPython.kernel", "-f", "{connection_file}"' in content


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# il faut toujours ceci au début du programme pour
# pouvoir mettre des accents dans le programme

# python ne fait pas grand chose mais il a plein de librairie
# en voici deux
import os
import pandas

###################
# retourner la matrice
###################

# changer False en True pour exécuter le code qui suit
if False:
    # on charge une feuille dans ce qu'on appelle un dataframe (chercher sur
    # wikipedia)
    df = pandas.read_excel("calories_a.xlsx")

    # on prend la transposée (les colonnes deviennent des lignes et
    # réciproquement)
    t = df.T

    # on stocke le résultat dans un autre fichier
    t.to_excel("calories_transposee.xlsx")

###################
# prendre plein de fichiers excel pour n'en faire qu'une seule table
###################

# changer False en True pour exécuter le code qui suit
if False:
    # données extraites depuis http://www.les-calories.com/calories-a.html

    # on récupère le contenu du répertoire
    fichiers = os.listdir(".")

    # on ne garde que les fichiers excel
    excel = [fichier for fichier in fichiers if "xlsx" in fichier]

    # on fait un print pour vérifier qu'on ne s'est pas trompé
    # normalement, tout s'affiche bien
    print(excel)

    # on ne garde que les fichiers qui ne contiennent pas transposee ni tout
    excel = [
        fichier for fichier in excel if "transposee" not in fichier and "tout" not in fichier]
    print(excel)

    # on charge tout les fichiers dans des dataframe
    liste = []
    for fichier in excel:
        print("lecture de ", fichier)
        df = pandas.read_excel(fichier)
        print("dimension", df.shape)
        liste.append(df)

    # on concatène le tout
    tout = pandas.concat(liste)
    print("dimension du tout", tout.shape)

    # on sauve le tout dans un unique fichier
    tout.to_excel("tout.xlsx")



"""
@brief      test log(time=3s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper import add_shortcut_to_desktop_for_module


class TestOnlyMain (unittest.TestCase):

    def test_spyder(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        if __name__ == "__main__":
            add_shortcut_to_desktop_for_module("spyder")


if __name__ == "__main__":
    unittest.main()



import sys
print(sys.executable)
import platform
print(platform.architecture())
sys.path.append("src")
print("----", sys.version_info)
from pymyinstall import win_python_setup
from pymyinstall.packaged import all_set, minimal_set, get_package_set


if False:
    from pymyinstall.win_installer.win_setup_r import _script
    r = r"C:\github\pymyinstall\dist\win_python_setup\tools\R"
    out = r_run_script(r, _script)
    print(out)


if False:
    list_modules = all_set()
    win_python_setup(module_list=list_modules, verbose=True,
                     download_only=False, no_setup=True,
                     selection={"R", "VS", "jenkins"},
                     source="2")

elif False:
    win_python_setup(module_list=minimal_set(), verbose=True,
                     download_only=False, no_setup=True,
                     selection={}, notebooks=[], source="2")
else:
    win_python_setup(module_list=get_package_set("datascientistbase")(),
                     verbose=True, download_only=False, no_setup=True,
                     selection={}, notebooks=[], source="2", embed=False)



# -*- coding: utf-8 -*-
"""
@file
@brief Implements command line ``python -m pymyinstall <args>``.
"""


if __name__ == "__main__":
    from pymyinstall.cli.pymy_install import main
    main()  # pragma: no cover



"""
@brief      test log(time=32s)
"""
import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installhelper.install_cmd_helper import run_cmd


class TestPyMyInstallCliTool(unittest.TestCase):

    def test_install_tool(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if not sys.platform.startswith("win"):
            # No installation on linux.
            return
        temp = get_temp_folder(__file__, "temp_install_tool")
        this = os.path.abspath(os.path.dirname(__file__))
        script = os.path.normpath(os.path.join(
            this, "..", "..", "src", "pymyinstall", "cli", "pymy_install.py"))
        cmd = "{0} -u {1} {2} --force --folder={3}".format(
            sys.executable, script, "graphviz --task=tool --source=zip", temp)
        out, err = run_cmd(cmd, wait=True)
        fLOG("----", cmd)
        fLOG(out.replace("\r", "").replace("\n\n", "\n"))
        fLOG("-----")
        fLOG(err.replace("\r", "").replace("\n\n", "\n"))
        content = os.listdir(temp)
        if not content:
            comp = "OUT:\n{0}\nERR:\n{1}".format(out, err)
            if sys.platform.startswith("win"):
                raise Exception("content is empty for: " + temp + "\n" + comp)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=2s)
"""
import unittest
import pandas
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pandashelper import df2rst
from pymyinstall.packaged import small_set, classifiers2string


class TestDocumentation(unittest.TestCase):

    def test_documentation(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mod = small_set()
        mod.sort()
        df = pandas.DataFrame(_.as_dict(rst_link=True) for _ in mod)
        df = df[["usage", "rst_link", "kind", "version",
                 "license", "purpose", "classifier"]]
        df["classifier"] = df.apply(
            lambda row: classifiers2string(row["classifier"]), axis=1)
        df.columns = ["usage", "name", "kind", "version",
                      "license", "purpose", "classifier"]
        fLOG(df2rst(df))


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=2s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.packaged import find_module_install, minimal_set, small_set


class TestDisributionSubSet(unittest.TestCase):

    def walk_text(self, set_mod):
        nb = 0
        for mod in set_mod:
            assert mod
            nb += 1
            find = find_module_install(mod.name, must_exist=True)
            assert find
        assert nb > 0
        return nb

    def test_small_set(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        nb = self.walk_text(small_set())
        fLOG(self._testMethodName, nb)

    def test_minimal_set(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        nb = self.walk_text(minimal_set())
        fLOG(self._testMethodName, nb)


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Various functions to install `InkScape <https://inkscape.org/>`_.
"""
from __future__ import print_function
import sys
import os
import re
from ..installhelper.install_cmd_helper import run_cmd
from .install_custom import download_page, download_file


def install_inkscape(dest_folder=".", fLOG=print, install=True, version=None):
    """
    Installs `InkScape <https://inkscape.org/en/>`_ (only on Windows).

    @param      dest_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      version         version to install (unused)
    @return                     temporary file

    """
    if version is not None:
        raise ValueError("cannot specify a version")
    if not sys.platform.startswith("win"):
        raise NotImplementedError(
            "MinGW can only be installed on Windows at the moment")

    link = "https://inkscape.org/en/download/windows/"
    page = download_page(link)
    if sys.platform.startswith("win"):
        reg = re.compile("href=\\\"(.*x64[.]msi)\\\"")
        alls = reg.findall(page)
        if len(alls) == 0:
            raise Exception(
                "unable to find a link on a .exe file on page: " +
                page)

        url = alls[0]
        outfile = os.path.join(dest_folder, url.split("/")[-1])
        if not os.path.exists(outfile):
            fLOG("[pymy] download ", url)
            local = download_file(url, outfile)
        else:
            return outfile
        if install:
            run_cmd("msiexec /i " + local, fLOG=fLOG, wait=True)
        return local
    else:
        raise NotImplementedError("not available on platform " + sys.platform)



"""
@brief      test log(time=2s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.packaged import small_set


class TestPackageListPrint(unittest.TestCase):

    def test_package_list_print(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        every = {}
        duplicated = []
        for _ in small_set():
            name = _.name
            if name in every:
                duplicated.append(name)
            every[name] = 1

        if len(duplicated) > 0:
            raise Exception("Duplicated modules\n{0}".format(
                "\n".join(duplicated)))

        fLOG(len(every))
        couples = [(k.lower(), k) for k in every]
        for _, k in sorted(couples):
            fLOG("'{0}',".format(k))


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Functions to prepare a setup on Windows, use InnoSetup
"""
from __future__ import print_function

import os
import sys
from ..installhelper.install_cmd_helper import run_cmd
from .win_ipy_kernels import install_kernels

if sys.version_info[0] == 2:
    from codecs import open
    FileNotFoundError = Exception


class InnoSetupException(Exception):

    """
    Exception happening with InnoSetup
    """
    pass


def find_innosetup():
    """
    find InnoSetup executable

    @return     executable
    """
    exe = r"C:\Program Files (x86)\Inno Setup 5\compil32.exe"
    if not os.path.exists(exe):
        raise FileNotFoundError(exe)
    return exe


def run_innosetup(new_script, innosetup=None, log_script=None, temp_folder=".", fLOG=print):
    """
    run InnotSetup for a script

    @param  new_script      script to run
    @param  innosetup       location of InnoSetup (if None, use default location)
    @param  log_script      output logs to this file
    @param  temp_folder     where to copy the modified script
    @param  fLOG            logging function
    @return                 output
    """
    if innosetup is None:
        innosetup = find_innosetup()

    cmd = [innosetup, "/cc", new_script]
    if log_script is not None:
        raise NotImplementedError()
        # cmd.append('/LOG="{0}"'.format(log_script))

    fLOG("[pymy] ISS script", new_script)
    fLOG("[pymy] CMD", cmd)
    out, err = run_cmd(" ".join(cmd), wait=True, fLOG=fLOG)
    if err is not None and len(err) > 0:
        raise InnoSetupException(
            "CMD:\n{0}\nOUT:\n{1}\nERR-5:\n{2}".format(cmd, out, err))
    return out


def innosetup_replacements(script=None, innosetup=None, replacements=None, log_script=None, temp_folder=".", fLOG=print):
    """
    run InnotSetup for a script

    @param  script          script to run, if None, use the default script assuming you want to build a Python Distribution
    @param  innosetup       location of InnoSetup (if None, use default location)
    @param  replacements    replace to make in the script (dictionary)
    @param  log_script      output logs to this file
    @param  temp_folder     where to copy the modified script
    @param  fLOG            logging function
    @return                 new script
    """
    if script is None:
        script = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "innosetup_script.iss"))

    if replacements is None:
        replacements = dict()

    with open(script, "r", encoding="utf8") as f:
        content = f.read()

    for k, v in replacements.items():
        content = content.replace(k, v)

    new_script = os.path.join(os.path.abspath(temp_folder),
                              os.path.split(script)[-1].replace(".iss", ".temp.iss"))
    with open(new_script, "w", encoding="utf8") as f:
        f.write(content)

    return new_script


def inno_install_kernels(root, suffix):
    """
    install kernels for Jupyter notebooks

    @param      root        root folder
    @param      suffix      suffix
    """
    if root in os.environ:
        path = os.environ[root]
        tools = os.path.join(path, "tools")
        if not os.path.exists(tools):
            tools = os.path.normpath(os.path.join(path, "..", "tools"))
            python = os.path.normpath(os.path.join(path, "..", "python"))
        else:
            python = os.path.join(path, "python")

    else:
        tools = os.path.join(root, "tools")
        python = os.path.join(root, "python")

    if not os.path.exists(tools):
        raise FileNotFoundError(tools)
    if not os.path.exists(python):
        raise FileNotFoundError(python)

    install_kernels(tools, python, suffix)



"""
@brief      test log(time=75s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installhelper.module_install import ModuleInstall


class TestDownloadPip (unittest.TestCase):

    def test_install_pip(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        if sys.version_info[0] == 2:
            # disabled on python 2.7
            return
        fold = get_temp_folder(__file__, "temp_download_pip")
        # This test can fail if the latest version of pyquickhelper is not
        # available for Python 2.7
        m = ModuleInstall("pyquickhelper", "pip", fLOG=fLOG)
        out = m.download(temp_folder=fold)
        self.assertTrue(os.path.exists(out))
        self.assertTrue("pyquickhelper" in out)
        out = m.download(temp_folder=fold)
        self.assertTrue(os.path.exists(out))
        self.assertTrue("pyquickhelper" in out)

    def test_install_pip_deps(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        if sys.version_info[0] == 2:
            # disabled on python 2.7
            return
        fold = get_temp_folder(__file__, "temp_download_pip_deps")
        # This test can fail if the latest version of pyquickhelper is not
        # available for Python 2.7
        m = ModuleInstall("pyquickhelper", "pip", fLOG=fLOG)
        out = m.download(temp_folder=fold, deps=True)
        self.assertTrue(os.path.exists(out))
        self.assertTrue("pyquickhelper" in out)
        out = m.download(temp_folder=fold, source="2")
        self.assertTrue(os.path.exists(out))
        self.assertTrue("pyquickhelper" in out)


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Various functions to install `MinGW <http://www.mingw.org/>`_.
"""
from __future__ import print_function
import sys
import os

from .install_custom import download_from_sourceforge


def install_mingw(dest_folder=".", fLOG=print, install=True, version=None):
    """
    install `MinGW <http://www.mingw.org/>`_ (only on Windows)

    @param      dest_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      version         version to install (unused)
    @return                     temporary file

    Packages to install:
        * gcc
        * g77 3.4.5
        * binutils
        * mingw-runtime
        * w32api
        * gcc-core - C compiler
        * gcc-g++ - C++ compiler
        * gcc-objc - Objective C compiler
        * gcc-gfortran - Fortran 90/95 compiler
        * gcc-java - Java compiler
        * gcc-ada - Ada compiler
        * mingw-gdb - Windows native build of GNU debugger
        * mingw32-make - Windows native build of GNU make
        * mingw-utils - Miscellaneous utilities

    ::

        mingw-get install binutils gcc g++ mingw32 fortran gdb mingw32 mingw w32api g77==3.4.5

    """
    if version is not None:
        raise ValueError("cannot specify a version")
    if not sys.platform.startswith("win"):
        raise NotImplementedError(
            "MinGW can only be installed on Windows at the moment")

    name = "mingw-get-setup.exe"
    newurl = "http://sourceforge.net/projects/mingw/files/Installer/{0}/download?use_mirror=autoselect".format(
        name)
    outfile = os.path.join(dest_folder, name)
    fLOG("[pymy] mingw, download from ", newurl)
    file = download_from_sourceforge(
        newurl,
        outfile,
        fLOG=fLOG,
        temp_folder=dest_folder)

    if install:
        raise NotImplementedError()
    return file



"""
@brief      test log(time=0s)
"""
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.filehelper import explore_folder_iterfile
from pyquickhelper.ipythonhelper import upgrade_notebook, remove_execution_number


class TestConvertNotebooks(unittest.TestCase):

    def test_convert_notebooks(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        fold = os.path.abspath(os.path.dirname(__file__))
        fold2 = os.path.normpath(
            os.path.join(fold, "..", "..", "_doc", "notebooks"))
        for nbf in explore_folder_iterfile(fold2, pattern=".*[.]ipynb"):
            t = upgrade_notebook(nbf)
            if t:
                fLOG("modified", nbf)
            # remove numbers
            remove_execution_number(nbf, nbf)

        fold2 = os.path.normpath(os.path.join(fold, "..", "..", "_unittests"))
        for nbf in explore_folder_iterfile(fold2, pattern=".*[.]ipynb"):
            if "site-packages" in nbf:
                continue
            t = upgrade_notebook(nbf)
            if t:
                fLOG("modified", nbf)


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Various functions to install `Java JDK <http://www.oracle.com/technetwork/java/javase/downloads/jdk8-downloads-2133151.html>`_
"""
from __future__ import print_function
import sys
import re
import os

from ..installhelper.install_cmd_helper import run_cmd
from .install_custom import download_page, download_file
from .install_custom_exceptions import ManualDownloadException


def install_javajdk(
        temp_folder=".", fLOG=print, install=True, force_download=False, version=None):
    """
    Install `Java JDK <http://www.oracle.com/technetwork/java/javase/downloads/jdk8-downloads-2133151.html>`_.

    @param      temp_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      force_download  force the downloading of pandoc
    @param      version         version to install (unused)
    @return                     temporary file
    """
    if version is not None:
        raise ValueError("cannot specify a version")
    #bb = IsPandocInstalled()
    bb = False
    if bb and not force_download:
        return True

    link = "http://www.oracle.com/technetwork/java/javase/downloads/jdk8-downloads-2133151.html"
    page = download_page(link)
    if sys.platform.startswith("win"):
        reg = re.compile("(http.*?jdk.*?windows-x64[.]exe)")
        alls = reg.findall(page)
        if len(alls) == 0:
            raise Exception(
                "unable to find a link on a .exe file on page: " +
                page)

        file = alls[-1].split("/")[-1]
        version = file.split("-")[1]
        b18 = int(version.replace("8u", "")) + 18 - 66
        root = "http://download.oracle.com/otn-pub/java/jdk/{0}-b{1}/".format(
            version, b18)

        filel = root + file
        fLOG("[pymy] java-jdk, version ", version)
        vershort = version.split("-")[0]
        full = filel.format(vershort, version)
        outfile = os.path.join(temp_folder, full.split("/")[-1])
        fLOG("[pymy] download ", full)
        local = download_file(full, outfile)
        size = os.stat(local).st_size
        if size < 2 ** 20:
            raise ManualDownloadException(
                "You should directly download the file from:\n{0}".format(link))
        if install and not bb:
            run_cmd("msiexec /i " + local, fLOG=fLOG, wait=True)
        return local
    else:
        raise NotImplementedError("not available on platform " + sys.platform)



"""
@file
@brief Functions to prepare a setup on Windows, R functions
"""
from __future__ import print_function

import os
import sys
from ..installhelper.install_cmd_helper import run_cmd

if sys.version_info[0] == 2:
    from codecs import open
    FileNotFoundError = Exception


_script_install = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "Julia_install.jl")

_script_build = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "Julia_build.jl")

_script_init = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "Julia_init.jl")


class JuliaBatchException(Exception):

    """
    raised when running R in batch mode
    """
    pass


def julia_run_script(julia_path, python_path, script, verbose=False, fLOG=print):
    """
    run a script on Julia

    @param      julia_path      julia location
    @param      script          script to run
    @param      python_path     path to python
    @param      verbose         more information
    @param      fLOG            logging function
    @return                     output
    """
    memo_path = os.environ["PATH"]
    epath = memo_path + ";" + \
        ";".join([python_path, os.path.join(python_path, "Scripts")])
    os.environ["PATH"] = epath

    exe = os.path.join(julia_path, "bin", "julia.exe")
    if not os.path.exists(exe):
        raise FileNotFoundError(exe)
    pkg = os.path.join(julia_path, "pkg")
    if not os.path.exists(pkg):
        os.mkdir(pkg)

    os.environ["JULIA_PKGDIR"] = pkg
    cmd = [exe, script, "--no_history-file"]
    cmd = " ".join(cmd)
    if verbose:
        fLOG("[pymy] set JULIA_PKGDIR=" + pkg)
    out, err = run_cmd(cmd, wait=True)
    if err is not None and len(err) > 0 and \
            "err" in err.lower() or "warn" in err.lower():
        raise JuliaBatchException(
            "CMD:\n{0}\nOUT:\n{1}\nERR--D:\n{2}".format(cmd, out, err))

    os.environ["PATH"] = memo_path
    patch_julia03(julia_path, verbose=verbose, fLOG=fLOG)
    return out


def patch_julia03(julia_path, verbose=False, fLOG=print):
    """
    patch absolute path in packages such as Julia/ZMQ or Julia/Nettle.

    @param      julia_path      julia_path
    @param      verbose         more information
    @param      fLOG            logging function
    """
    pkg = os.path.join(julia_path, "pkg")
    pkg_d = pkg.replace("\\", "\\\\")
    if verbose:
        fLOG("[pymy]   string to replace", pkg_d)
    for root, dirs, files in os.walk(pkg):
        for name in files:
            if name.endswith("deps.jl"):
                full = os.path.join(root, name)
                with open(full, "r", encoding="utf8") as f:
                    content = f.read()
                if pkg_d in content:
                    content = content.replace(pkg_d, "%JULIA_PKGDIR%")
                    if verbose:
                        fLOG("[pymy]   patch ", full)
                    with open(full, "w", encoding="utf8") as f:
                        f.write(content)



"""
@file
@brief Helpers for function @see fn win_python_setup
"""
from __future__ import print_function
import datetime
import os
import shutil
import sys
import fnmatch
import subprocess

from ..installcustom.install_custom_pandoc import install_pandoc
from ..installcustom.install_custom_R import install_R
from ..installcustom.install_custom_julia import install_julia
from ..installcustom.install_custom_scite import install_scite
from ..installcustom.install_custom_putty import install_putty
from ..installcustom.install_custom_sqlitespy import install_sqlitespy
from ..installcustom.install_custom_python import install_python
from ..installcustom.install_custom_mingw import install_mingw
from ..installcustom.install_custom_tdm_gcc import install_tdm_gcc
from ..installcustom.install_custom_7z import install_7z
from ..installcustom.install_custom_graphviz import install_graphviz
from ..installcustom.install_custom_vs import install_vs
from ..installcustom.install_custom import download_page
from ..installcustom.install_custom_javajdk import install_javajdk
from ..installcustom.install_custom_jenkins import install_jenkins
from ..installcustom.install_custom_git import install_git
from ..installcustom.install_custom_miktex import install_miktex
from ..installcustom.install_custom_inkscape import install_inkscape
from ..installhelper.link_shortcuts import add_shortcut
from ..installhelper import run_cmd
from ..packaged import minimal_set

from .win_packages import _is_package_in_list
from .pywin32_helper import import_pywin32
from .win_extract import extract_msi, extract_exe, extract_archive, extract_copy
from .win_exception import WinInstallException
from .win_setup_mark_step import mark_step, is_step_done
from .win_setup_r import r_run_script, _script as _script_r
from .win_setup_julia import julia_run_script, _script_install as _script_julia_install
from .win_setup_julia import _script_build as _script_julia_build, _script_init as _script_julia_init


if sys.version_info[0] == 2:
    from codecs import open
    FileNotFoundError = Exception


def dtnow():
    """
    shortcut, return ``datetime.datetime.now()``
    """
    return datetime.datetime.now()


def copy_icons(src, dest):
    """
    copy all files from src to dest

    @param      src     source
    @param      dest    destination
    @return             operations
    """
    operations = []
    if not os.path.exists(dest):
        os.makedirs(dest)
        operations.append(("mkdir", dest))
    files = os.listdir(src)
    for file in files:
        if os.path.isdir(os.path.join(src, file)):
            continue
        d = os.path.join(dest, file)
        if not os.path.exists(d):
            shutil.copy(os.path.join(src, file), dest)
            operations.append(("copy", file))
    return operations


def win_download(folder=None, module_list=None, verbose=False, fLOG=print,
                 download_only=True, selection=None, source=None, embed=True):
    """
    The function downloads everything needed to prepare a setup.

    @param      folder          where to prepare the python version (the user must replace None)
    @param      module_list     list of module to install (see @see fn minimal_set = default options)
    @param      fLOG            logging function
    @param      download_only   only downloads (unused, True by default)
    @param      verbose         print more information
    @param      selection       selection of tools to install (dictionary { toolname: version or None})
    @param      source          source of python packages (see @see cl ModuleInstall)
    @param      embed           use embedded version (or custom one)
    @return                     list of completed operations

    List of available tools:

    * scite
    * putty
    * mingw
    * SQLiteSpy
    * python
    * R
    * vs
    * julia
    * 7z
    * graphviz
    * tdm
    * jdk (java)
    * jenkins
    * git
    * miktex
    * inkscape
    """
    if selection is None:
        raise ValueError("selection must be specified")

    available = os.listdir(folder)

    def is_here(program, no_wheel=False):
        b = _is_package_in_list(program, available, no_wheel=no_wheel)
        return b

    operations = []

    if not is_here("scite") and "scite" in selection:
        if verbose:
            fLOG("[pymy] --- download", "scite")
        r = install_scite(dest_folder=folder, fLOG=fLOG,
                          install=False, version=selection.get("scite", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("putty") and "putty" in selection:
        if verbose:
            fLOG("[pymy] --- download", "putty")
        r = install_putty(dest_folder=folder, fLOG=fLOG,
                          install=False, version=selection.get("putty", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("mingw") and "mingw" in selection:
        if verbose:
            fLOG("[pymy] --- download", "mingw")
        r = install_mingw(dest_folder=folder, fLOG=fLOG,
                          install=False, version=selection.get("mingw", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("miktex") and "miktex" in selection:
        if verbose:
            fLOG("[pymy] --- download", "miktex")
        r = install_miktex(dest_folder=folder, fLOG=fLOG,
                           install=False, version=selection.get("miktex", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("inkscape") and "inkscape" in selection:
        if verbose:
            fLOG("[pymy] --- download", "inkscape")
        r = install_inkscape(dest_folder=folder, fLOG=fLOG,
                             install=False, version=selection.get("inkscape", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("tdm") and "tdm" in selection:
        if verbose:
            fLOG("[pymy] --- download", "tdm")
        r = install_tdm_gcc(dest_folder=folder, fLOG=fLOG,
                            install=False, version=selection.get("tdm", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("git") and "git" in selection:
        if verbose:
            fLOG("[pymy] --- download", "git")
        r = install_git(folder, fLOG=fLOG,
                        install=False, version=selection.get("git", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("SQLiteSpy") and "sqlitespy" in selection:
        if verbose:
            fLOG("[pymy] --- download", "sqllitespy")
        r = install_sqlitespy(temp_folder=folder, fLOG=fLOG,
                              install=False, version=selection.get("sqlitespy", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("python") and "python" in selection:
        if verbose:
            fLOG("[pymy] --- download", "python")
        r = install_python(
            temp_folder=folder, fLOG=fLOG, install=False, force_download=True,
            version=selection.get("python", None), custom=not embed)
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("R-") and "r" in selection:
        if verbose:
            fLOG("[pymy] --- download", "R")
        r = install_R(
            temp_folder=folder, fLOG=fLOG, install=False, force_download=True, version=selection.get("r", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("vs") and "vs" in selection:
        if verbose:
            fLOG("[pymy] --- download", "Visual Studio Express")
        r = install_vs(folder, fLOG=fLOG, install=False,
                       version=selection.get("vs", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("julia") and "julia" in selection:
        if verbose:
            fLOG("[pymy] --- download", "julia")
        r = install_julia(
            temp_folder=folder, fLOG=fLOG, install=False, force_download=True, version=selection.get("julia", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("pandoc") and "pandoc" in selection:
        if verbose:
            fLOG("[pymy] --- download", "pandoc")
        r = install_pandoc(
            temp_folder=folder, fLOG=fLOG, install=False, force_download=True, version=selection.get("pandoc", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("7z") and "7z" in selection:
        if verbose:
            fLOG("[pymy] --- download", "7z")
        r = install_7z(
            temp_folder=folder, fLOG=fLOG, install=False, force_download=True, version=selection.get("7z", None))
        operations.append(("download", r))
        fLOG("done")

    if not is_here("graphviz", no_wheel=True) and "graphviz" in selection:
        if verbose:
            fLOG("[pymy] --- download", "graphviz")
        r = install_graphviz(
            temp_folder=folder, fLOG=fLOG, install=False, force_download=True, version=selection.get("graphviz", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("jdk", no_wheel=True) and "jdk" in selection:
        if verbose:
            fLOG("[pymy] --- download", "java jdk")
        r = install_javajdk(
            temp_folder=folder, fLOG=fLOG, install=False, force_download=True, version=selection.get("java jdk", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if not is_here("jenkins", no_wheel=True) and "jenkins" in selection:
        if verbose:
            fLOG("[pymy] --- download", "jenkins")
        r = install_jenkins(folder, fLOG=fLOG, install=False,
                            version=selection.get("jenkins", None))
        operations.append(("download", r))
        fLOG("[pymy] done")

    if module_list is None:
        module_list = minimal_set()

    for mod in module_list:
        if is_here(mod.name + "-"):
            continue
        if verbose:
            fLOG("[pymy] download module", mod.name)
        res = mod.download(temp_folder=folder, source=source)
        if isinstance(res, list):
            for r in res:
                operations.append(("download", r))
        else:
            operations.append(("download", res))

    return operations


def win_install(folders, download_folder, verbose=False, fLOG=print,
                names="Julia Scite 7z TDM MinGW R pandoc Python SQLiteSpy Putty Graphviz".split(),
                selection=None):
    """
    Install setups

    @param      folders         dictionary of folders, must contain key tools, python
    @param      download_folder where the setup are
    @param      fLOG            logging function
    @param      verbose         print more information
    @param      names           name of subfolders to be created
    @param      selection       selection of tools to install
    @return                     list of completed operations, executable (to make shortcuts)

    The function installs every setup which starts by one of the string in *names*
    and whose extension is .exe, .msi, .zip, .7z.

    To install Python on Windows,
    see `Using Python on Windows <https://docs.python.org/3.5/using/windows.html>`_,
    and `What's coming for the Python 3.5 installer? <http://stevedower.id.au/blog/the-python-3-5-installer/>`_
    """
    operations = []
    dfunc = {".zip": extract_archive, ".exe": extract_exe,
             ".msi": extract_msi, "putty.exe": extract_copy,
             ".7z": extract_archive}

    def location(file):
        ext = os.path.splitext(file)[-1]
        if ext not in [".msi", ".exe", ".zip", ".7z"]:
            return None
        lf = file.lower()
        for name in names:
            lname = name.lower()
            if name.lower() in selection and (lf.startswith(lname) or "-%s-" % lname in lf):
                if name == "Python":
                    return folders["python"]
                else:
                    return os.path.join(folders["tools"], name)
        return None

    def find_exe(loc, name):
        name = name.lower()
        if name.lower() == "mingw":
            name = "gcc"
        elif name.lower() == "tdm":
            name = "gcc"
        elif name.lower() == "jdk":
            name = "java"
        elif name.lower() == "graphviz":
            name = "dot"
        exp = name + ".exe"

        for root, dirnames, filenames in os.walk(loc):
            for filename in fnmatch.filter(filenames, '*.exe'):
                exe = os.path.join(root, filename)
                file = os.path.split(exe)[-1].lower()
                if file == exp:
                    return exe
        return None

    # we sort to get 7z installed first as it is needed for exe files
    cands = os.listdir(download_folder)
    cands.sort()

    installed = {}

    for cand in cands:
        loc = location(cand)
        if loc is None:
            continue

        name = os.path.split(loc)[-1]
        if not os.path.exists(loc):
            os.mkdir(loc)

        # already installed, checking if there is any exe file
        exe = find_exe(loc, name)
        if exe is not None:
            # already done
            fLOG("[pymy] --- already installed", exe)
        else:
            fLOG("[pymy] --- install", cand, " in ", loc)
            full = os.path.join(download_folder, cand)
            if 'tdm' in cand and 'gcc' in cand:
                raise WinInstallException(
                    "TM must be manually installed from the setup\n{0}\nin\n{1}".format(full, loc))
            elif 'python' in cand:
                continue
                # see http://stevedower.id.au/blog/the-python-3-5-installer/
                temploc = os.path.join(download_folder, "python")
                options = [os.path.join(
                    download_folder, cand), "/quiet", "/layout", temploc]
                cmd = " ".join(options)
                fLOG("[pymy] run ", cmd)
                out, err = run_cmd(cmd, wait=True)
                fLOG("[pymy] OUT:\n", out)
                if err:
                    fLOG("[pymy] OUT:\n", err)
                options = [os.path.join(temploc, cand),
                           "TargetDir={0}".format(loc), "InstallAllUsers=1",
                           "AssociateFiles=0", "CompileAll=1", "Include_symbols=1",
                           "SimpleInstall=1"]
                cmd = " ".join(options)
                fLOG("[pymy] run ", cmd)
                out, err = run_cmd(cmd, wait=True)
                fLOG("[pymy] OUT:\n", out)
                if err:
                    fLOG("[pymy] OUT:\n", err)
            else:
                ext = os.path.splitext(cand)[-1]
                filename = os.path.split(cand)[-1]
                func = dfunc.get(filename, dfunc[ext])

                if ext == ".exe":
                    func(
                        full, loc, verbose=verbose, fLOG=fLOG, szip=installed["7z"])
                else:
                    func(full, loc, verbose=verbose, fLOG=fLOG)

            operations.append(("install", cand))
            fLOG("[pymy] done")

        # add main executable
        found = find_exe(loc, name)

        # for MinGW, we check that the executable mingw-get.exe was installed
        if found is None:
            if name == "MinGW" and cand == "mingw-get-setup.exe":
                exe = os.path.join(loc, "bin", "mingw-get.exe")
                if os.path.exists(exe):
                    cmd = exe + \
                        " install binutils gcc g++ mingw32 fortran gdb mingw32 mingw w32api g77"
                    if verbose:
                        fLOG("[pymy] install MinGW", cmd)
                    retcode = subprocess.call(
                        cmd, shell=True, stdout=sys.stderr)
                    if retcode < 0:
                        raise WinInstallException(
                            "unable to execute:\nCMD:\n{0}".format(cmd))
                    found = find_exe(loc, name)
            elif name == "VS" and cand == "vs_community.exe":
                # Visual Studio needs to be manually installed
                # we copy the setup to VS
                shutil.copy(os.path.join(download_folder, cand), loc)
                found = os.path.join(loc, cand)

        if found is None:
            raise FileNotFoundError("unable to find executable for name={0} in {1}, found: {2}, exe={3}, function={4}".format(
                name, loc, found, exe, func))
        installed[name] = found

    return operations, installed


def create_links_tools(folder, installed, verbose=False, fLOG=print):
    """
    create links for the tools

    @param      folder      where links will be stored
    @param      installed   dictionary *{ tool: exe file }*
    @param      verbose     display more information
    @param      fLOG        logging function
    @return                 operations, list of tuple *("link", link file)*
    """
    import_pywin32()
    operations = []

    for k, v in installed.items():
        if k == "R":
            name = "test R Gui"
            link_name = name + ".lnk"
            dest = os.path.join(folder, link_name)
            add_shortcut(target="tools\\R\\bin\\x64\\Rgui.exe",
                         name=name, arguments="", icon="~dp0\\tools\\icons\\r.ico",
                         folder=folder)
            if verbose:
                fLOG("[pymy] create link", dest)
            operations.append(("link", link_name))

            name = "test R Console"
            link_name = name + ".lnk"
            dest = os.path.join(folder, link_name)
            add_shortcut(target="tools\\R\\bin\\x64\\R.exe",
                         name=name, arguments="", icon="~dp0\\tools\\icons\\r.ico",
                         folder=folder)
            if verbose:
                fLOG("[pymy] create link", dest)
            operations.append(("link", link_name))

        elif k == "julia":
            name = "test Julia Console"
            link_name = name + ".lnk"
            dest = os.path.join(folder, link_name)
            add_shortcut(target="tools\\julia\\bin\\julia.exe",
                         name=name, arguments="", icon="~dp0\\tools\\icons\\julia.ico",
                         folder=folder)
            if verbose:
                fLOG("[pymy] create link", dest)
            operations.append(("link", link_name))

        elif k == "python":
            name = "test Python Console"
            link_name = name + ".lnk"
            dest = os.path.join(folder, link_name)
            add_shortcut(target="python\\python.exe",
                         name=name, arguments="", icon="~dp0\\tools\\icons\\python.ico",
                         folder=folder)
            if verbose:
                fLOG("[pymy] create link", dest)
            operations.append(("link", link_name))

    return operations


def win_download_notebooks(notebooks, folder, verbose=False, fLOG=print):
    """
    download notebooks and store them as documentation

    @param      notebooks       list of tuple (place, url)
    @param      folder          where to put them
    @param      verbose         verbose
    @param      fLOG            logging function
    @return                     list of operations
    """
    operations = []
    for path, urls in notebooks:
        if not isinstance(urls, list):
            urls = [urls]
        for url in urls:
            name = url.split("/")[-1]
            dest = os.path.join(folder, path)
            if not os.path.exists(dest):
                os.makedirs(dest)
                operations.append(("mkdir", dest))
            dfile = os.path.join(dest, name)
            if not os.path.exists(dfile):
                if verbose:
                    fLOG("download notebooks", name, "from", url)
                content = download_page(url)
                with open(dfile, "w", encoding="utf8") as f:
                    f.write(content)
                operations.append(("docs", dfile))
    return operations


def win_install_julia_step(folders, verbose=False, fLOG=print):
    """
    does necessary steps to setup Julia

    @param      folders     installation folders
    @param      verbose     verbose
    @param      fLOG        logging function
    @return                 list of processed operations
    """
    operations = []
    if not is_step_done(folders["logs"], "julia_init"):
        ##########################
        # init Julia packages
        #########################
        fLOG("[pymy] --- init julia packages")
        jl = os.path.join(folders["tools"], "Julia")
        output = os.path.join(folders["logs"], "out.init.julia.txt")
        out = julia_run_script(
            jl, folders["python"], _script_julia_init, verbose=verbose, fLOG=fLOG)
        with open(output, "w", encoding="utf8") as f:
            f.write(out)
        operations.append(("Julia", _script_julia_init))
        operations.append(("time", dtnow()))
        mark_step(folders["logs"], "julia_init")

    if not is_step_done(folders["logs"], "julia_install"):
        ##########################
        # install Julia packages
        #########################
        fLOG("[pymy] --- install julia packages")
        jl = os.path.join(folders["tools"], "Julia")
        output = os.path.join(folders["logs"], "out.install.julia.txt")
        out = julia_run_script(
            jl, folders["python"], _script_julia_install, verbose=verbose, fLOG=fLOG)
        with open(output, "w", encoding="utf8") as f:
            f.write(out)
        operations.append(("Julia", _script_julia_install))
        operations.append(("time", dtnow()))
        mark_step(folders["logs"], "julia_install")

    if not is_step_done(folders["logs"], "julia_build"):
        #########################
        # build Julia packages
        #########################
        fLOG("[pymy] --- build julia packages")
        jl = os.path.join(folders["tools"], "Julia")
        output = os.path.join(folders["logs"], "out.build.julia.txt")
        out = julia_run_script(
            jl, folders["python"], _script_julia_build, verbose=verbose, fLOG=fLOG)
        with open(output, "w", encoding="utf8") as f:
            f.write(out)
        operations.append(("Julia", _script_julia_build))
        operations.append(("time", dtnow()))
        mark_step(folders["logs"], "julia_build")

    return operations


def win_install_r_step(folders, verbose=False, fLOG=print):
    """
    does necessary steps to setup R

    @param      folders     installation folders
    @param      verbose     verbose
    @param      fLOG        logging function
    @return                 list of processed operations
    """
    operations = []
    if not is_step_done(folders["logs"], "r_install"):
        ######################
        # install R packages
        ######################
        fLOG("[pymy] --- install R packages")
        r = os.path.join(folders["tools"], "R")
        output = os.path.join(folders["logs"], "out.install.r.txt")
        out = r_run_script(r, _script_r, output)
        with open(output, "w", encoding="utf8") as f:
            f.write(out)
        operations.append(("R", _script_r))
        operations.append(("time", dtnow()))
        mark_step(folders["logs"], "r_install")
    else:
        fLOG(
            "--- skip installation of R packages or remove file log.step.r_install.txt")

    return operations



"""
@brief      test log(time=3s)
"""

import sys
import os
import unittest
import warnings
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.win_installer.win_ipython_helper import ipython_create_profile, ipython_update_profile
from pymyinstall.installhelper.run_cmd import RunCmdException


class TestIPythonProfile(unittest.TestCase):

    def test_ipython_profile(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        python_path = os.path.abspath(os.path.dirname(sys.executable))
        temp = get_temp_folder(__file__, "temp_ipython_profile")
        try:
            path = ipython_create_profile(
                temp, python_path, name="ZZZ", fLOG=fLOG)
        except (FileNotFoundError, RunCmdException) as e:
            if "_venv" in str(e) and (".exe" in str(e) or "/_venv/" in str(e)):
                # ipython.exe not present in virtual environment
                warnings.warn("needs to add custom command line")
                return
            else:
                raise Exception(
                    "Error message\n---\n{0}\n---".format(e)) from e

        fLOG(path)
        assert os.path.exists(path)

        if sys.platform.startswith("win"):
            ipython_update_profile(path)
            profile = os.path.join(path, "ipython_kernel_config.py")
            with open(profile, "r", encoding="utf8") as f:
                lines = f.readlines()

            exp = "c.ContentsManager.hide_globs = ['__pycache__',"
            nb = 0
            for line in lines:
                if line.startswith(exp):
                    nb += 1
            assert nb > 0
        else:
            profile = os.path.join(path, "ipython_config.py")
            assert os.path.exists(profile)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
@brief      test log(time=25s)
"""
import os
import unittest
import warnings
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder, ExtTestCase
from pymyinstall.setuphelper import create_folder_setup


class TestSetupFolder(ExtTestCase):

    def test_setup_folder(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        temp = get_temp_folder(__file__, "temp_setup_folder", clean=False)

        try:
            import cairocffi
        except (ImportError, OSError) as e:
            warnings.warn(str(e))
            return
        self.assertTrue(cairocffi is not None)
        st = create_folder_setup('cairocffi', fLOG=fLOG, output_path=temp)
        self.assertEqual(len(st), 1)
        self.assertExists(st[0])
        exp = os.path.join(temp, 'dist')
        found = os.listdir(exp)
        self.assertEqual(len(found), 1)
        self.assertIn(".whl", found[0])


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Various functions to install `SQLiteSpy <http://www.yunqa.de/delphi/doku.php/products/sqlitespy/index>`_.
"""
from __future__ import print_function
import sys
import re
import os

from ..installhelper.install_cmd_helper import unzip_files
from ..installhelper.link_shortcuts import add_shortcut_to_desktop
from .install_custom import download_page, download_from_sourceforge, DownloadException, download_file


def IsSQLiteSpyInstalled(dest_folder):
    """
    Checks if :epkg:`SQLiteSpy` was already installed

    @param      dest_folder     where it was installed
    @return                     boolean
    """
    if sys.platform.startswith("win"):
        file = os.path.join(dest_folder, "SQLiteSpy.exe")
        return os.path.exists(file)
    else:
        raise NotImplementedError("not available on platform " + sys.platform)


def install_sqlitespy(temp_folder=".", fLOG=print, install=True, version=None, backup=False):
    """
    Installs :epkg:`SQLiteSpy`.
    It does not do it a second time if it is already installed.

    @param      temp_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      version         version to install (unused)
    @return                     temporary file
    """
    if version is not None:
        raise ValueError("cannot specify a version")
    if IsSQLiteSpyInstalled(temp_folder):
        return os.path.join(temp_folder, "SQLiteSpy.exe")

    link = "https://www.yunqa.de/delphi/products/sqlitespy/index"
    outfile = None
    try:
        page = download_page(link)
    except DownloadException as e:
        if backup and sys.platform.startswith("win"):
            link = "http://www.xavierdupre.fr/enseignement/setup/SQLiteSpy_1.9.12.zip"
            fLOG("[pymy] download ", link)
            outfile = os.path.join(temp_folder, "SQLiteSpy_1.9.12.zip")
            download_file(link, outfile)
            version = "1.9.12"
            fLOG("[pymy] SQLiteSpy, version ", version)
        else:
            raise e

    if sys.platform.startswith("win"):
        if outfile is None:
            reg = re.compile(
                "href=[\\\"'](https://www.yunqa.de/delphi/downloads/SQLiteSpy.*?[.]zip)[\\\"']")
            alls = reg.findall(page)
            if len(alls) == 0:
                raise Exception(
                    "Unable to find a link on a .zip file on page: " + page)

            file = alls[0].replace("&amp;", "&")
            full = file
            version = file.split("_")[-1].replace(".zip", "")
            fLOG("[pymy] SQLiteSpy, version ", version)
            outfile = os.path.join(
                temp_folder, "{0}_{1}.zip".format("SQLiteSpy", version))
            fLOG("[pymy] download ", full)
            download_from_sourceforge(
                full, outfile, temp_folder=temp_folder, fLOG=fLOG)
        if install:
            files = unzip_files(outfile, temp_folder, fLOG=fLOG)
            local = [f for f in files if f.endswith(".exe")][0]
            return local
        else:
            return outfile
    else:
        raise NotImplementedError("not available on platform " + sys.platform)


def add_shortcut_to_desktop_for_sqlitespy(exe):
    """
    create a shortcut on your desktop

    @param      exe        exe location (SQLiteSpy.exe)
    @return                filename
    """
    return add_shortcut_to_desktop(exe, "SQLiteSpy", "SQLiteSpy")



"""
@file
@brief Shortuts
"""
from .install_cmd_helper import run_cmd, update_pip, has_pip, add_shortcut_to_desktop_for_module, get_pip_program
from .install_venv_helper import run_cmd_path, run_venv_script, venv_install, create_virtual_env
from .module_install import ModuleInstall
from .module_dependencies import missing_dependencies
from .module_install_version import get_module_dependencies, get_module_version, get_pypi_version, get_module_metadata
from .module_install_version import version_consensus, numeric_version, compare_version, is_installed, get_wheel_version
from .status_helper import get_installed_modules


def module_as_table(list_module, as_df=False):
    """
    Returns a list of dictionaries or a dataframe
    for a list of modules.

    @param      list_module     list of @see cl ModuleInstall
    @param      as_df           as a dataframe or not
    @return                     list of dictionaries or dataframe
    """
    res = [_.as_dict() for _ in list_module]
    if as_df:
        import pandas
        return pandas.DataFrame(res)
    else:
        return res



"""
@brief      test log(time=45s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installcustom import install_putty


class TestPutty (unittest.TestCase):

    def test_install_putty(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        fold = os.path.abspath(os.path.split(__file__)[0])
        temp = os.path.join(fold, "temp_putty")
        if not os.path.exists(temp):
            os.mkdir(temp)
        for _ in os.listdir(temp):
            f = os.path.join(temp, _)
            if os.path.isfile(f):
                os.remove(f)

        if sys.platform.startswith("win"):
            r = install_putty(temp, fLOG=fLOG)
            exe = os.path.abspath(r)
            fLOG(exe)
            self.assertTrue(os.path.exists(exe))


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Various functions to install some application such as `pandoc <http://johnmacfarlane.net/pandoc/>`_.
"""
from __future__ import print_function
import os
import sys
from ..installhelper.install_memoize import install_memoize2
from ..installhelper.internet_settings import default_user_agent

if sys.version_info[0] == 2:
    FileNotFoundError = Exception
    import urllib2 as urllib_request
    import urllib2 as urllib_error
    import httplib as http_client
else:
    import urllib.request as urllib_request
    import urllib.error as urllib_error
    import http.client as http_client


class DownloadException(Exception):
    """
    Cannot download.
    """
    pass


@install_memoize2
def download_page(url, is406=False):
    """
    Downloads a page from a url.

    @param      url     url
    @param      is406   if the function raised *HTTP Error 406*, try *True*
    @return             content

    .. versionchanged:: 1.1
        Parameter *is406* was added.
    """
    agent = "Mozilla" if is406 else default_user_agent
    try:
        req = urllib_request.Request(
            url, headers={
                'User-agent': agent})
        u = urllib_request.urlopen(req)
        text = u.read()
        u.close()
    except urllib_error.HTTPError as e:
        raise DownloadException(
            "Unable to get archive from: '" + url + "'.") from e
    except urllib_error.URLError as e:
        raise DownloadException(
            "Unable to get archive from: '" + url + "'.") from e
    except ConnectionResetError as e:
        raise DownloadException(
            "Unable to get archive from: '" + url + "'.") from e

    typstr = str  # unicode#
    return typstr(text, encoding="utf8")


def download_file(url, outfile, fLOG=None):
    """
    Downloads a file from a url, the function does not download the file
    again if outfile already exists.

    @param      url         url
    @param      outfile     outfile
    @param      fLOG      logging function
    @return                 outfile
    """
    if os.path.exists(outfile):
        return outfile

    try:
        if fLOG:
            fLOG("[pymy] download", url)
        req = urllib_request.Request(
            url,
            headers={
                'User-agent': default_user_agent},
        )
        u = urllib_request.urlopen(req)
        text = u.read()
        u.close()
    except urllib_error.HTTPError as e:
        raise DownloadException(
            "Unable to get archive from '{}'.".format(url)) from e
    except urllib_error.URLError as e:
        raise DownloadException(
            "Unable to get archive from '{}'.".format(url)) from e
    except http_client.IncompleteRead as ee:
        raise DownloadException(
            "unable to complete reading from: " + url) from ee

    with open(outfile, "wb") as f:
        f.write(text)

    return outfile


def download_from_sourceforge(url, outfile, fLOG=print, temp_folder="."):
    """
    Downloads a file from a url using redirection,
    the function does not download the file
    again if outfile already exists.

    @param      url         url
    @param      outfile     outfile
    @param      fLOG        logging function
    @param      temp_folder only used if installation of module requests is needed
    @return                 outfile

    The function will install module `requests <http://docs.python-requests.org/en/latest/>`_
    if not present.
    """
    if os.path.exists(outfile):
        return outfile

    try:
        import requests
    except ImportError:
        fLOG("[pymy] installing module requests")
        from ..installhelper.module_install import ModuleInstall
        ModuleInstall("requests", fLOG=fLOG).install(temp_folder=temp_folder)
        import requests

    try:
        req = requests.get(url, allow_redirects=True, stream=True)
        text = req.raw.read()
        fLOG("[pymy] len ", len(text))
    except urllib_error.HTTPError as e:
        raise DownloadException("unable to get archive from: " + url) from e
    except requests.exceptions.ConnectionError as ee:
        raise DownloadException("unable to get archive from: " + url) from ee

    if len(text) < 20 and text.decode(
            "ascii").lower().startswith("bad request"):
        raise Exception("Bad Request for url: " + url)

    with open(outfile, "wb") as f:
        f.write(text)

    return outfile


def where_in_path(name):
    """
    Looks for a file in current directory and in every path in ``PATH``.

    @param      name        name of the file to look for
    @return                 None if not found, the absolute filename otherwise
    """
    if os.path.exists(name):
        return os.path.abspath(name)
    else:
        path = os.environ["PATH"]
        if path:
            spl = path.split(";")
            for p in spl:
                new_name = os.path.join(p, name)
                if os.path.exists(new_name):
                    return new_name
    return None



"""
@file
@brief Helpers for virtualenv
"""
from __future__ import print_function

import os
import sys
from .install_cmd_helper import run_cmd

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


class VirtualEnvError(Exception):
    """
    exception raised by the function implemented in this file
    """
    pass


def build_venv_cmd(params, posparams):
    """
    builds the command line for virtual env

    @param      params      dictionary of parameters
    @param      posparams   positional arguments
    @return                 string
    """
    import venv
    dir(venv)
    exe = sys.executable.replace("w.exe", "").replace(".exe", "")
    cmd = [exe, "-m", "venv"]
    for k, v in params.items():
        if v is None:
            cmd.append("--" + k)
        else:
            cmd.append("--" + k + "=" + v)
    cmd.extend(posparams)
    return " ".join(cmd)


def create_virtual_env(where, symlinks=False, system_site_packages=False,
                       clear=True, packages=None, fLOG=print,
                       temp_folder=None):
    """
    .. index:: virtual environment

    create a virtual environment

    @param      where                   location of this virtual environment
    @param      symlinks                attempt to symlink rather than copy
    @param      system_site_packages    Give the virtual environment access to the system site-packages dir
    @param      clear                   Delete the environment directory if it already exists.
                                        If not specified and the directory exists, an error is raised.
    @param      packages                list of packages to install (it will install module
                                        :epkg:`pymyinstall`).
    @param      fLOG                    logging function
    @param      temp_folder             temporary folder (to download module if needed), by default ``<where>/download``
    @return                             stand output

    .. exref::
        :title: Create a virtual environment

        The following example creates a virtual environment.
        Packages can be added by specifying the parameter *package*.

        ::

            from pyquickhelper.pycode import create_virtual_env
            fold = "my_env"
            if not os.path.exists(fold):
                os.mkdir(fold)
            create_virtual_env(fold)
    """
    fLOG("[pymy] create virtual environment at:", where)
    params = {}
    if symlinks:
        params["symlinks"] = None
    if system_site_packages:
        params["system-site-packages"] = None
    if clear:
        params["clear"] = None
    cmd = build_venv_cmd(params, [where])
    out, err = run_cmd(cmd, wait=True, fLOG=fLOG)
    if len(err) > 0:
        raise VirtualEnvError(
            "unable to create virtual environement at {2}\nCMD:\n{3}\nOUT:\n{0}\nERR-G:\n{1}".format(out, err, where, cmd))

    if sys.platform.startswith("win"):
        scripts = os.path.join(where, "Scripts")
    else:
        scripts = os.path.join(where, "bin")

    if not os.path.exists(scripts):
        files = "\n  ".join(os.listdir(where))
        raise FileNotFoundError(
            "unable to find {0}, content:\n  {1}".format(scripts, files))

    in_scripts = os.listdir(scripts)
    pips = [_ for _ in in_scripts if _.startswith("pip")]
    if len(pips) == 0:
        out += venv_install(where, "pip", fLOG=fLOG,
                            temp_folder=temp_folder)
    in_scripts = os.listdir(scripts)
    pips = [_ for _ in in_scripts if _.startswith("pip")]
    if len(pips) == 0:
        raise FileNotFoundError(
            "unable to find pip in {0}, content:\n  {1}".format(scripts, in_scripts))

    if packages is not None and len(packages) > 0:
        fLOG("[pymy] install packages in:", where)
        packages = [_ for _ in packages if _ != "pymyinstall" and _ != "pip"]
        if len(packages) > 0:
            out += venv_install(where, packages, fLOG=fLOG,
                                temp_folder=temp_folder)
    return out


def venv_install(venv, packages, fLOG=print, temp_folder=None):
    """
    install a package or a list of packages in a virtual environment

    @param      venv            location of the virtual environment
    @param      packages        a package (str) or a list of packages(list[str])
    @param      fLOG            logging function
    @param      temp_folder     temporary folder (to download module if needed), by default ``<where>/download``
    @return                     standard output
    """
    if temp_folder is None:
        temp_folder = os.path.join(venv, "download")

    if isinstance(packages, str):
        packages = [packages]

    if packages == "pip" or packages == ["pip"]:
        from .get_pip import __file__ as pip_loc
        ppath = os.path.abspath(pip_loc.replace(".pyc", ".py"))
        script = ["-u", ppath]
        out = run_venv_script(venv, script, fLOG=fLOG, is_cmd=True)

        if sys.platform.startswith("win"):
            scripts = os.path.join(venv, "Scripts")
        else:
            scripts = os.path.join(venv, "bin")
        in_scripts = os.listdir(scripts)
        pips = [_ for _ in in_scripts if _.startswith("pip")]
        if len(pips) == 0:
            raise FileNotFoundError(
                "unable to find pip in {0},\nvenv:\n{2}\nppath:\n{3}\ncontent:\n  {1}".format(scripts, in_scripts, venv, ppath))
        return out
    else:
        p = os.path.normpath(os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "..", ".."))
        ls = ','.join("'{0}'".format(_) for _ in packages)
        script = ["import sys",
                  "sys.path.append('{0}')".format(p.replace("\\", "\\\\")),
                  "import pymyinstall",
                  "ps=[{0}]".format(ls),
                  "t='{0}'".format(temp_folder.replace("\\", "\\\\")),
                  "pymyinstall.packaged.install_all(temp_folder=t,list_module=ps,up_pip=False)"]
        return run_venv_script(venv, "\n".join(script), fLOG=fLOG)


def run_venv_script(venv, script, fLOG=print, file=False, is_cmd=False):
    """
    run a script on a virtual environment (the script should be simple

    @param      venv        virtual environment
    @param      script      script as a string (not a file)
    @param      fLOG        logging function
    @param      file        is script a file or a string to execute
    @param      is_cmd      if True, script is a command line to run (as a list) for python executable
    @return                 output
    """
    if sys.platform.startswith("win"):
        exe = os.path.join(venv, "Scripts", "python")
    else:
        exe = os.path.join(venv, "bin", "python")
    if is_cmd:
        cmd = " ".join([exe, script])
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG)
        if len(err) > 0:
            raise VirtualEnvError(
                "unable to run cmd at {2}\nCMD:\n{3}\nOUT:\n{0}\nERR-H:\n{1}".format(out, err, venv, cmd))
        return out
    else:
        script = ";".join(script.split("\n"))
        if file:
            if not os.path.exists(script):
                raise FileNotFoundError(script)
            cmd = " ".join([exe, "-u", '"{0}"'.format(script)])
        else:
            cmd = " ".join([exe, "-u", "-c", '"{0}"'.format(script)])
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG)
        if len(err) > 0:
            raise VirtualEnvError(
                "unable to run script at {2}\nCMD:\n{3}\nOUT:\n{0}\nERR-I:\n{1}".format(out, err, venv, cmd))
        return out


def run_cmd_path(python_path, script, fLOG=print, file=False, is_cmd=False, **kwargs):
    """
    run a script knowing python path, it does not raise an exception.

    @param      python_path     python path
    @param      script          script as a string (not a file) or command line if *is_cmd* is True
    @param      fLOG            logging function
    @param      file            is script a file or a string to execute
    @param      is_cmd          if True, script is a command line to run (as a list) for python executable
    @param      kwargs          extra parameters
    @return                     output, error
    """
    if sys.platform.startswith("win"):
        exe = os.path.join(python_path, "python")
    else:
        python_name = "python" if sys.version_info[
            0] == 2 else "python%d" % sys.version_info[0]
        exe = os.path.join(python_path, python_name)
        if not os.path.exists(exe):
            exe = os.path.join(python_path, "bin", python_name)
    if is_cmd:
        cmd = " ".join([exe] + script)
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG, **kwargs)
        return out, err
    else:
        script = ";".join(script.split("\n"))
        if file:
            if not os.path.exists(script):
                raise FileNotFoundError(script)
            cmd = " ".join([exe, "-u", '"{0}"'.format(script)])
        else:
            cmd = " ".join([exe, "-u", "-c", '"{0}"'.format(script)])
        out, err = run_cmd(cmd, wait=True, fLOG=fLOG, **kwargs)
        return out, err



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.module_install_version import choose_most_recent


class TestInstallModuleVersionFast(unittest.TestCase):

    def test_choose_most_recent(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mods = [("**", 'numpy-1.10.0+mkl-cp34-none-win_amd64.whl'),
                ("**", 'numpy-1.9.3+mkl-cp34-none-win_amd64.whl')]
        ch = choose_most_recent(mods)
        fLOG(ch)
        self.assertEqual(
            ch, ('**', 'numpy-1.10.0+mkl-cp34-none-win_amd64.whl'))


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Various functions to install `reveal.js <https://github.com/hakimel/reveal.js/releases>`_.
"""
from __future__ import print_function
import re
import os
import sys

from .install_custom import download_page, download_file
from ..installhelper.install_cmd_helper import unzip_files

if sys.version_info[0] == 2:
    FileNotFoundError = Exception


def download_revealjs(
        temp_folder=".",
        unzip_to=".",
        fLOG=print,
        force_download=False,
        location="https://github.com/hakimel/reveal.js/releases",
        clean=True,
        version=None):
    """
    Download `reveal.js <https://github.com/hakimel/reveal.js/releases>`_ release.
    and unzip it.

    @param      temp_folder     where to download the setup
    @param      unzip_to        where to unzip the files
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      force_download  force the downloading of pandoc
    @param      location        location of reveal.js release
    @param      clean           clean unnecessary files
    @param      version         version to download (unused)
    @return                     list of downloaded and unzipped files
    """
    if version is not None:
        raise ValueError("cannot specify a version")
    link = location
    page = download_page(link)
    reg = re.compile("href=\\\"(.*?[.]zip)\\\"")
    alls = reg.findall(page)
    if len(alls) == 0:
        raise Exception(
            "unable to find a link on a .zip file on page: " +
            page)

    filename = alls[0].split("/")[-1]
    filel = location.replace("releases", "").rstrip(
        "/") + "/archive/{0}".format(filename)
    outfile = os.path.join(temp_folder, "reveal.js." + filename)
    version = ".".join(filel.split("/")[-1].split(".")[:-1])
    fLOG("[pymy] download ", filel, "to", outfile, "version", version)
    local = download_file(filel, outfile)
    res = unzip_files(local, whereTo=unzip_to, fLOG=fLOG)

    master = os.path.join(unzip_to, "reveal.js-%s" % version)
    if not os.path.exists(master):
        raise FileNotFoundError("unable to find: " + master)
    new_master = os.path.join(unzip_to, "reveal.js")

    os.rename(master, new_master)
    res = [r.replace(master, new_master) for r in res]

    keep = []
    for r in res:
        if os.path.isdir(r):
            continue
        if ".gitignore" in r or ".travis.yml" in r or "index.html" in r \
           or "README" in r or "CONTRIBUTING.md" in r:
            os.remove(r)
        elif "/test/" in r.replace("\\", "/"):
            os.remove(r)
        else:
            keep.append(r)

    return keep



"""
@file
@brief Guess missing dependencies
"""
from .pip_helper import get_installed_distributions


def get_default_ignore_modules():
    """
    A couple of modules have some dependencies on not supported modules
    or modules integrated to the standard distributiion such
    `enum34 <https://pypi.python.org/pypi/enum34>`_.
    This function returns this list.
    Skips errors such as ``Module 'cachecontrol' misses 'msgpack_python'``.
    """
    return [
        "argparse",          # standard distribution (>= 3.5)
        "backports.weakref",  # standard distribution (>= 3.4)
        "distribute",        # standard distribution
        "enum",              # standard distribution (>= 3.4)
        "enum34",            # standard distribution (>= 3.4)
        "enum_compat",       # standard distribution (>= 3.4)
        "enum-compat",       # standard distribution (>= 3.4)
        "fasttsne",          # deprecated module
        "futures",           # standard distribution (>= 3.5)
        "holoviews",         # needed by geoviews but still marked as missing
        "hypertemp",         # temporary solution, should disappear
        "keyrings.alt",      # weird this one on Windows
        # "guidata",
        # "monotonic",
        "msgpack_python",   # Called msgpack now.
        "msgpack-python",   # Called msgpack now.
        "onnx",             # working on it
        "ordereddict",      # standard distribution (>= 3.4)
        "path",             # subfolder of existing module
        "path.py",          # weird
        "pathlib",          # standard distribution (>= 3.5)
        "pip",              # almost standard distribution, should be rare
        "pipdeptree",       # weird
        # manny names for this one (PTable, not maintained anymore)
        "prettytable",
        "pycryptodome",     # many names for this one
        "pydocstyle",       # does not seem to be maintained
        "pyopengl",         # weird this one
        "pypiwin32",        # manny names for this one
        "pythonnet",        # weird
        "pywin32",          # manny names for this one
        "pywin32_ctypes",   # manny names for this one
        "rope",             # for spyder
        "setuptools",       # standard distribution
        "typing",           # standard distribution (>= 3.6)
    ]


def _main_pipdeptree(local_only=False):
    """
    The function relies on module
    `pipdeptree.main <https://pypi.python.org/pypi/pipdeptree>`_.
    """
    default_skip = ['setuptools', 'pip', 'python', 'distribute', 'hypertemp']
    skip = default_skip + ['pipdeptree']
    pkgs = get_installed_distributions(local_only=local_only, skip=skip)
    req_map = dict((p.key, (p, p.requires())) for p in pkgs)
    return req_map


def missing_dependencies(specific=None, ignore_module=get_default_ignore_modules()):
    """
    Returns the list of missing dependencies for the installed modules.

    @param      specific        look dependencies only for a specific module
    @param      ignore_module   list of modules not to consider as a missing dependency
                                even if they are installed
    @return                     list of missing dependencies as dictionary (module, missing dependencies)

    .. versionchanged:: 1.5
        Parameters *ignore_module*.
    """
    skip = set(ignore_module)
    tree = _main_pipdeptree()
    stack = {}
    for k, v in tree.items():
        if specific is not None and specific != k:
            continue
        for mod in v[1]:
            dep = mod.key
            if dep in skip:
                continue
            if dep not in tree:
                if "-" in dep:
                    dep = dep.replace("-", "_")
                elif "_" in dep:
                    dep = dep.replace("_", "_")
            if dep not in tree:
                if k not in stack:
                    stack[k] = []
                stack[k].append(dep)
    return stack



"""
@file
@brief Various functions to install some applications such as `7z <http://www.7-zip.org/>`_.
"""
from __future__ import print_function
import sys
import re
import os

from ..installhelper.install_cmd_helper import run_cmd
from .install_custom import download_page, download_file


def install_7z(
        temp_folder=".", fLOG=print, install=True, force_download=False, version=None):
    """
    Install `7z <http://www.7-zip.org/>`_.
    It does not do it a second time if it is already installed.

    @param      temp_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      force_download  force the downloading of 7z
    @param      version         specify a version (unused)
    @return                     temporary file
    """
    if version is not None:
        raise ValueError("cannot specify a version")
    link = "http://www.7-zip.org/download.html"
    page = download_page(link)
    if sys.platform.startswith("win"):
        reg = re.compile("href=\\\"(.*?x64[.]msi)\\\"")
        alls = reg.findall(page)
        if len(alls) == 0:
            raise Exception(
                "unable to find a link on a .msi file on page: " + link + "\n" +
                page)

        url = "/".join(link.split("/")[:-1]) + "/" + alls[0]
        full = url.split("/")[-1]
        outfile = os.path.join(temp_folder, full)
        local = download_file(url, outfile, fLOG=fLOG)
        if install:
            run_cmd("msiexec /i " + local, fLOG=fLOG, wait=True)
        return local
    else:
        raise NotImplementedError("not available on platform " + sys.platform)



"""
@file
@brief Some fixes for deprecated functions and not updated packages.
"""
import os
import sys


def fix_win32ctypes_core_cffi__advapi32_py(filename=None, fileout=None, fLOG=None):
    """
    Fixes the following issue:

    ::

          File "C:\\Python37-x64\\lib\\site-packages\\win32ctypes\\pywin32\\__init__.py", line 21, in <module>
            from win32ctypes.pywin32 import win32api
          File "C:\\Python37-x64\\lib\\site-packages\\win32ctypes\\pywin32\\win32api.py", line 23, in <module>
            from win32ctypes.core import _common, _kernel32, _backend
          File "C:\\Python37-x64\\lib\\site-packages\\win32ctypes\\core\\__init__.py", line 39, in <module>
            from .cffi import _advapi32, _common, _kernel32
          File "C:\\Python37-x64\\lib\\site-packages\\win32ctypes\\core\\cffi\\_advapi32.py", line 198

            ^
        SyntaxError: invalid syntax
        Command exited with code 1

    If that errors happens, you could add:

    ::

        python -c "from pymyinstall.fix import fix_win32ctypes_core_cffi__advapi32_py;fix_win32ctypes_core_cffi__advapi32_py(fLOG=print)"
    """
    if sys.platform.startswith("win") and sys.version_info[:2] >= (3, 6):
        if filename is None:
            filename = os.path.join(os.path.dirname(
                sys.executable), "lib\\site-packages\\win32ctypes\\core\\cffi\\_advapi32.py")
        if not os.path.exists(filename):
            if fLOG:
                fLOG(
                    "[fix_win32ctypes_core_cffi__advapi32_py] '{0}' not found".format(filename))
            return
        if fileout is None:
            fileout = filename
        if fLOG:
            fLOG(
                "[fix_win32ctypes_core_cffi__advapi32_py] found '{0}'".format(filename))
        with open(filename, "r", encoding="utf-8") as f:
            content = f.read()
        content = content.replace("\r", "\n").replace("\n\n", "\n")
        with open(fileout, "w", encoding="utf-8") as f:
            f.write(content)


if __name__ == "__main__":
    fix_win32ctypes_core_cffi__advapi32_py()



"""
@file
@brief Various to set up a C++ compiler
"""
import os


def switch_to_VS_compiler(python_path, version=12):
    """
    applies fix described in `Build a Python 64 bit extension on Windows 8 <http://www.xavierdupre.fr/blog/2013-07-07_nojs.html>`_

    @param  python_path     python path
    @param  version         Visual Studio version, 12 = VS2013
    @return                 impacted files
    """
    name = os.path.join(python_path, "Lib", "distutils", "msvc9compiler.py")
    with open(name, "r") as f:
        content = f.read()
    content = content.replace(
        "'win-amd64' : 'amd64',", "'win-amd64' : 'x86_amd64',")
    content = content.replace(
        "if majorVersion >= 6:", "if majorVersion >= 6 :\n        majorVersion = " + str(version))
    with open(name, "w") as f:
        f.write(content)
    return [name]


def switch_to_mingw_compiler(python_path):
    """
    applies a fix to use MinGW to compile extensions (does not work with Jupyter)

    @param  python_path     python path
    """
    dirname = os.path.join(python_path, "Lib", "distutils")
    cfg = os.path.join(dirname, "distutils.cfg")
    with open(cfg, "w") as f:
        f.write("""[build]
                    compiler = mingw32

                    [build_ext]
                    compiler = mingw32
                    """.replace("                            ", ""))
    return [cfg]
